#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2012,2020 Nick Hall
# Copyright (C) 2012      Rob G. Healey
# Copyright (C) 2012      Benny Malengier
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Gramps setuptools module.
"""

# check python version first
import sys

if sys.version_info < (3, 9):
    raise SystemExit("Gramps requires Python 3.9 or later.")

from setuptools import setup, Command

try:
    from setuptools.command.build import build as _build
except ImportError:
    from distutils.command.build import build as _build
import os
import glob
import codecs
import subprocess
from stat import ST_MODE
from gramps.version import VERSION
import unittest
import argparse
import logging

_LOG = logging.getLogger(".setup")

svem_flag = "--single-version-externally-managed"
if svem_flag in sys.argv:
    # Die, setuptools, die.
    sys.argv.remove(svem_flag)

argparser = argparse.ArgumentParser(add_help=False)
argparser.add_argument(
    "--no-compress-manpages", dest="no_compress_manpages", action="store_true"
)
args, passthrough = argparser.parse_known_args()
sys.argv = [sys.argv[0]] + passthrough


def newer(source, target):
    """
    Determines if a target file needs to be rebuilt.

    Returns True if the target file doesn't exist or if the source file is
    newer than the target file.
    """
    if not os.path.exists(target):
        return True
    return os.path.getmtime(source) > os.path.getmtime(target)


def substitute_variables(filename_in, filename_out, subst_vars):
    """
    Substitute variables in a file.
    """
    f_in = codecs.open(filename_in, encoding="utf-8")
    f_out = codecs.open(filename_out, encoding="utf-8", mode="w")
    for line in f_in:
        for variable, substitution in subst_vars:
            line = line.replace(variable, substitution)
        f_out.write(line)
    f_in.close()
    f_out.close()


def get_linguas():
    """
    Read the po/LINGUAS file to get a list of all linguas.
    """
    all_linguas = []
    with open(os.path.join("po", "LINGUAS"), "r", encoding="utf-8") as linguas:
        for line in linguas:
            if "#" in line:
                line = line[: line.find("#")]
            all_linguas.extend(line.split())
    return all_linguas


def build_trans(build_cmd):
    """
    Translate the language files into gramps.mo
    """
    data_files = build_cmd.distribution.data_files
    for lang in get_linguas():
        po_file = os.path.join("po", lang + ".po")
        mo_file = os.path.join(
            build_cmd.build_base, "mo", lang, "LC_MESSAGES", "gramps.mo"
        )
        mo_file_unix = build_cmd.build_base + "/mo/" + lang + "/LC_MESSAGES/gramps.mo"
        mo_dir = os.path.dirname(mo_file)
        if not (os.path.isdir(mo_dir) or os.path.islink(mo_dir)):
            os.makedirs(mo_dir)

        if newer(po_file, mo_file):
            cmd = ["msgfmt", po_file, "-o", mo_file]
            if subprocess.call(cmd) != 0:
                os.remove(mo_file)
                msg = "ERROR: Building language translation files failed."
                ask = msg + "\n Continue building y/n [n] "
                reply = input(ask)
                if reply in ["n", "N"]:
                    raise SystemExit(msg)
            _LOG.info("Compiling %s >> %s", po_file, mo_file)

        # linux specific piece:
        target = "share/locale/" + lang + "/LC_MESSAGES"
        data_files.append((target, [mo_file_unix]))


def build_man(build_cmd):
    """
    Compresses Gramps manual files
    """
    data_files = build_cmd.distribution.data_files
    for man_dir, dirs, files in os.walk(os.path.join("data", "man")):
        if "gramps.1.in" in files:
            filename = os.path.join(man_dir, "gramps.1.in")
            newdir = os.path.join(build_cmd.build_base, man_dir)
            if not (os.path.isdir(newdir) or os.path.islink(newdir)):
                os.makedirs(newdir)

            newfile = os.path.join(newdir, "gramps.1")
            subst_vars = (("@VERSION@", VERSION),)
            substitute_variables(filename, newfile, subst_vars)

            src = "gramps.1"
            if not args.no_compress_manpages:
                import gzip

                src += ".gz"
                man_file_gz = os.path.join(newdir, src)
                if os.path.exists(man_file_gz):
                    if newer(filename, man_file_gz):
                        os.remove(man_file_gz)
                    else:
                        filename = False
                        os.remove(newfile)

                if filename:
                    # Binary io, so open is OK
                    with open(newfile, "rb") as f_in, gzip.open(
                        man_file_gz, "wb"
                    ) as f_out:
                        f_out.writelines(f_in)
                        _LOG.info("Compiling %s >> %s", filename, man_file_gz)

                    os.remove(newfile)
                    filename = False

            lang = man_dir[8:]
            src = build_cmd.build_base + "/data/man" + lang + "/" + src
            target = "share/man" + lang + "/man1"
            data_files.append((target, [src]))


def build_intl(build_cmd):
    """
    Merge translation files into desktop and mime files
    """
    data_files = build_cmd.distribution.data_files
    base = build_cmd.build_base

    merge_files = (
        ("org.gramps_project.Gramps.desktop", "share/applications", "--desktop"),
        ("org.gramps_project.Gramps.xml", "share/mime/packages", "--xml"),
        ("org.gramps_project.Gramps.metainfo.xml", "share/metainfo", "--xml"),
    )

    for filename, target, option in merge_files:
        filenamelocal = os.path.join("data", filename)
        newfile = os.path.join(base, filenamelocal)
        newdir = os.path.dirname(newfile)
        if not (os.path.isdir(newdir) or os.path.islink(newdir)):
            os.makedirs(newdir)
        merge(filenamelocal + ".in", newfile, option)
        data_files.append((target, [base + "/data/" + filename]))


def merge(in_file, out_file, option, po_dir="po"):
    """
    Run the msgfmt command.
    """
    if not os.path.exists(out_file) and os.path.exists(in_file):
        cmd = ["msgfmt", option, "--template", in_file, "-d", po_dir, "-o", out_file]
        if subprocess.call(cmd, env={**os.environ, "GETTEXTDATADIR": po_dir}) != 0:
            msg = "ERROR: %s was not merged into the translation files!\n" % out_file
            raise SystemExit(msg)
        _LOG.info("Compiling %s >> %s", in_file, out_file)


class build(_build):
    """Custom build command."""

    def run(self):
        build_trans(self)
        if not sys.platform == "win32":
            build_man(self)
        build_intl(self)
        _build.run(self)


class test(Command):
    """Command to run Gramps unit tests"""

    description = "run all unit tests"
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        if not os.path.exists("build"):
            raise RuntimeError(
                "No build directory. Run `python setup.py build` before trying to run tests."
            )
        os.environ["GRAMPS_RESOURCES"] = "."
        all_tests = unittest.TestLoader().discover(".", pattern="*_test.py")
        result = unittest.TextTestRunner(verbosity=self.verbose).run(all_tests)
        if not result.wasSuccessful():
            raise SystemExit("ERROR: Unit test failure.")


# -------------------------------------------------------------------------
#
# Packages
#
# -------------------------------------------------------------------------
package_core = [
    "gramps",
    "gramps.cli",
    "gramps.cli.plug",
    "gramps.gen.utils.docgen",
    "gramps.gen",
    "gramps.gen.datehandler",
    "gramps.gen.db",
    "gramps.gen.db.conversion_tools",
    "gramps.gen.display",
    "gramps.gen.filters",
    "gramps.gen.filters.rules",
    "gramps.gen.filters.rules.citation",
    "gramps.gen.filters.rules.event",
    "gramps.gen.filters.rules.family",
    "gramps.gen.filters.rules.media",
    "gramps.gen.filters.rules.note",
    "gramps.gen.filters.rules.person",
    "gramps.gen.filters.rules.place",
    "gramps.gen.filters.rules.repository",
    "gramps.gen.filters.rules.source",
    "gramps.gen.lib",
    "gramps.gen.merge",
    "gramps.gen.mime",
    "gramps.gen.plug",
    "gramps.gen.plug.docbackend",
    "gramps.gen.plug.docgen",
    "gramps.gen.plug.menu",
    "gramps.gen.plug.report",
    "gramps.gen.proxy",
    "gramps.gen.simple",
    "gramps.gen.utils",
    "gramps.gen.utils.docgen",
    "gramps.test",
    "gramps.plugins",
    "gramps.plugins.cite",
    "gramps.plugins.db",
    "gramps.plugins.db.bsddb",
    "gramps.plugins.db.dbapi",
    "gramps.plugins.docgen",
    "gramps.plugins.drawreport",
    "gramps.plugins.export",
    "gramps.plugins.gramplet",
    "gramps.plugins.graph",
    "gramps.plugins.importer",
    "gramps.plugins.lib",
    "gramps.plugins.lib.maps",
    "gramps.plugins.mapservices",
    "gramps.plugins.quickview",
    "gramps.plugins.rel",
    "gramps.plugins.sidebar",
    "gramps.plugins.textreport",
    "gramps.plugins.thumbnailer",
    "gramps.plugins.tool",
    "gramps.plugins.view",
    "gramps.plugins.webreport",
    "gramps.plugins.webstuff",
]
package_gui = [
    "gramps.gui",
    "gramps.gui.editors",
    "gramps.gui.editors.displaytabs",
    "gramps.gui.filters",
    "gramps.gui.filters.sidebar",
    "gramps.gui.logger",
    "gramps.gui.merge",
    "gramps.gui.plug",
    "gramps.gui.plug.export",
    "gramps.gui.plug.quick",
    "gramps.gui.plug.report",
    "gramps.gui.selectors",
    "gramps.gui.views",
    "gramps.gui.views.treemodels",
    "gramps.gui.widgets",
]

packages = package_core + package_gui

# -------------------------------------------------------------------------
#
# Package data
#
# -------------------------------------------------------------------------

# add all subdirs of plugin with glade:
package_data_core = []
basedir = os.path.join("gramps", "plugins")
for dirpath, dirnames, filenames in os.walk(basedir):
    root, subdir = os.path.split(dirpath)
    if subdir.startswith("."):
        dirnames[:] = []
        continue
    for dirname in dirnames:
        # Skip hidden and system directories:
        if dirname.startswith("."):
            dirnames.remove(dirname)
        # we add to data_list so glade , xml, files are found, we don't need the gramps/ part
        package_data_core.append(dirpath[7:] + "/" + dirname + "/*.glade")
        package_data_core.append(dirpath[7:] + "/" + dirname + "/*.xml")
        package_data_core.append(dirpath[7:] + "/" + dirname + "/*.ini")

package_data_gui = ["gui/glade/*.glade"]
package_data = package_data_core + package_data_gui

# -------------------------------------------------------------------------
#
# Resources
#
# -------------------------------------------------------------------------
data_files_core = []
DOC_FILES = ["AUTHORS", "COPYING", "FAQ", "INSTALL", "NEWS", "README.md", "TODO"]
GEDCOM_FILES = glob.glob(os.path.join("example", "gedcom", "*.*"))
GRAMPS_FILES = glob.glob(os.path.join("example", "gramps", "*.*"))
IMAGE_WEB = glob.glob(os.path.join("images", "webstuff", "*.png"))
IMAGE_WEB.extend(glob.glob(os.path.join("images", "webstuff", "*.ico")))
IMAGE_WEB.extend(glob.glob(os.path.join("images", "webstuff", "*.gif")))
GRAMPS_CSS = glob.glob(os.path.join("data", "*.css"))
CSS_FILES = glob.glob(os.path.join("data", "css", "*.css"))
JS_FILES = glob.glob(os.path.join("data", "css", "*.js"))
SWANKY_PURSE = glob.glob(os.path.join("data", "css", "swanky-purse", "*.css"))
SWANKY_IMG = glob.glob(os.path.join("data", "css", "swanky-purse", "images", "*.png"))
data_files_core.append(("share/doc/gramps", DOC_FILES))
data_files_core.append(("share/doc/gramps/example/gedcom", GEDCOM_FILES))
data_files_core.append(("share/doc/gramps/example/gramps", GRAMPS_FILES))
data_files_core.append(("share/gramps/images/webstuff", IMAGE_WEB))
data_files_core.append(("share/gramps", GRAMPS_CSS))
data_files_core.append(("share/gramps/css", CSS_FILES + JS_FILES))
data_files_core.append(("share/gramps/css/swanky-purse", SWANKY_PURSE))
data_files_core.append(("share/gramps/css/swanky-purse/images", SWANKY_IMG))

DTD_FILES = glob.glob(os.path.join("data", "*.dtd"))
RNG_FILES = glob.glob(os.path.join("data", "*.rng"))
XML_FILES = glob.glob(os.path.join("data", "*.xml"))
data_files_core.append(("share/gramps", XML_FILES))
data_files_core.append(("share/gramps", DTD_FILES))
data_files_core.append(("share/gramps", RNG_FILES))

data_files_gui = []
IMAGE_FILES = glob.glob(os.path.join("images", "*.*"))
THEME = os.path.join("images", "hicolor")
ICON_16 = glob.glob(os.path.join(THEME, "16x16", "actions", "*.png"))
ICON_22 = glob.glob(os.path.join(THEME, "22x22", "actions", "*.png"))
ICON_24 = glob.glob(os.path.join(THEME, "24x24", "actions", "*.png"))
ICON_48 = glob.glob(os.path.join(THEME, "48x48", "actions", "*.png"))
ICON_SC = glob.glob(os.path.join(THEME, "scalable", "actions", "*.svg"))
data_files_gui.append(("share/gramps/images", IMAGE_FILES))
data_files_gui.append(("share/gramps/images/hicolor/16x16/actions", ICON_16))
data_files_gui.append(("share/gramps/images/hicolor/22x22/actions", ICON_22))
data_files_gui.append(("share/gramps/images/hicolor/24x24/actions", ICON_24))
data_files_gui.append(("share/gramps/images/hicolor/48x48/actions", ICON_48))
data_files_gui.append(("share/gramps/images/hicolor/scalable/actions", ICON_SC))
APP_16 = os.path.join(THEME, "16x16", "apps", "org.gramps_project.Gramps.png")
APP_22 = os.path.join(THEME, "22x22", "apps", "org.gramps_project.Gramps.png")
APP_24 = os.path.join(THEME, "24x24", "apps", "org.gramps_project.Gramps.png")
APP_48 = os.path.join(THEME, "48x48", "apps", "org.gramps_project.Gramps.png")
APP_128 = os.path.join(THEME, "128x128", "apps", "org.gramps_project.Gramps.png")
APP_256 = os.path.join(THEME, "256x256", "apps", "org.gramps_project.Gramps.png")
APP_SC = os.path.join(THEME, "scalable", "apps", "org.gramps_project.Gramps.svg")
data_files_gui.append(("share/icons/hicolor/16x16/apps", [APP_16]))
data_files_gui.append(("share/icons/hicolor/22x22/apps", [APP_22]))
data_files_gui.append(("share/icons/hicolor/24x24/apps", [APP_24]))
data_files_gui.append(("share/icons/hicolor/48x48/apps", [APP_48]))
data_files_gui.append(("share/icons/hicolor/128x128/apps", [APP_128]))
data_files_gui.append(("share/icons/hicolor/256x256/apps", [APP_256]))
data_files_gui.append(("share/icons/hicolor/scalable/apps", [APP_SC]))
MIME_48 = glob.glob(os.path.join(THEME, "48x48", "mimetypes", "*.png"))
MIME_SC = glob.glob(os.path.join(THEME, "scalable", "mimetypes", "*.svg"))
data_files_gui.append(("share/icons/hicolor/48x48/mimetypes", MIME_48))
data_files_gui.append(("share/icons/hicolor/scalable/mimetypes", MIME_SC))

data_files = data_files_core + data_files_gui

with open("README.md", "r") as f:
    readme_text = f.read()

# -------------------------------------------------------------------------
#
# Setup
#
# -------------------------------------------------------------------------
setup(
    name="gramps",
    description="Gramps (Genealogical Research and Analysis Management Programming System)",
    long_description=readme_text,
    long_description_content_type="text/markdown",
    version=VERSION,
    author="Donald N. Allingham",
    author_email="don@gramps-project.org",
    maintainer="Gramps Development Team",
    maintainer_email="benny.malengier@gmail.com",
    url="http://gramps-project.org",
    license="GPL v2 or greater",
    platforms=["FreeBSD", "Linux", "MacOS", "Windows"],
    cmdclass={"build": build, "test": test},
    packages=packages,
    package_data={"gramps": package_data},
    install_requires=["orjson"],
    extras_require={
        "image": ["Pillow"],
        "i18n": ["PyICU"],
        "GUI": ["PyGObject", "pycairo"],
        "all": [
            "Pillow",
            "PyGObject",
            "PyICU",
            "imagesize",
            "pycairo",
            "pycountry",
        ],
        "testing": ["jsonschema", "mock", "lxml"],
    },
    data_files=data_files,
    scripts=["scripts/gramps"],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Environment :: MacOS X",
        "Environment :: Plugins",
        "Environment :: Web Environment",
        "Environment :: Win32 (MS Windows)",
        "Environment :: X11 Applications :: GTK",
        "Intended Audience :: Education",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: Other Audience",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        "Natural Language :: Arabic",
        "Natural Language :: Bulgarian",
        "Natural Language :: Catalan",
        "Natural Language :: Chinese (Simplified)",
        "Natural Language :: Croatian",
        "Natural Language :: Czech",
        "Natural Language :: Danish",
        "Natural Language :: Dutch",
        "Natural Language :: English",
        "Natural Language :: Esperanto",
        "Natural Language :: Finnish",
        "Natural Language :: French",
        "Natural Language :: German",
        "Natural Language :: Greek",
        "Natural Language :: Hebrew",
        "Natural Language :: Hungarian",
        "Natural Language :: Icelandic",
        "Natural Language :: Italian",
        "Natural Language :: Japanese",
        "Natural Language :: Macedonian",
        "Natural Language :: Norwegian",
        "Natural Language :: Polish",
        "Natural Language :: Portuguese",
        "Natural Language :: Portuguese (Brazilian)",
        "Natural Language :: Romanian",
        "Natural Language :: Russian",
        "Natural Language :: Serbian",
        "Natural Language :: Slovak",
        "Natural Language :: Slovenian",
        "Natural Language :: Spanish",
        "Natural Language :: Swedish",
        "Natural Language :: Turkish",
        "Natural Language :: Vietnamese",
        "Operating System :: MacOS",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: Other OS",
        "Operating System :: POSIX :: BSD",
        "Operating System :: POSIX :: Linux",
        "Operating System :: POSIX :: SunOS/Solaris",
        "Operating System :: Unix",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Topic :: Database",
        "Topic :: Desktop Environment :: Gnome",
        "Topic :: Education",
        "Topic :: Scientific/Engineering :: Visualization",
        "Topic :: Sociology :: Genealogy",
    ],
)



# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2004-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Swedish-specific classes for parsing and displaying dates.
"""

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import re

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..lib.date import Date
from ._dateparser import DateParser
from ._datedisplay import DateDisplay
from ._datehandler import register_datehandler


# -------------------------------------------------------------------------
#
# Swedish parser class
#
# -------------------------------------------------------------------------
class DateParserSv(DateParser):
    """
    Convert a text string into a Date object, expecting a date
    notation in the swedish language. If the date cannot be converted,
    the text string is assigned.
    """

    # modifiers before the date
    modifier_to_int = {
        "före": Date.MOD_BEFORE,
        "innan": Date.MOD_BEFORE,
        "efter": Date.MOD_AFTER,
        "omkring": Date.MOD_ABOUT,
        "ca": Date.MOD_ABOUT,
        "c:a": Date.MOD_ABOUT,
        "från": Date.MOD_FROM,
        "till": Date.MOD_TO,
    }

    bce = ["f Kr"]

    calendar_to_int = {
        "gregoriansk   ": Date.CAL_GREGORIAN,
        "g": Date.CAL_GREGORIAN,
        "juliansk": Date.CAL_JULIAN,
        "j": Date.CAL_JULIAN,
        "hebreisk": Date.CAL_HEBREW,
        "h": Date.CAL_HEBREW,
        "islamisk": Date.CAL_ISLAMIC,
        "muslimsk": Date.CAL_ISLAMIC,
        "i": Date.CAL_ISLAMIC,
        "fransk": Date.CAL_FRENCH,
        "fransk republikansk": Date.CAL_FRENCH,
        "f": Date.CAL_FRENCH,
        "persisk": Date.CAL_PERSIAN,
        "p": Date.CAL_PERSIAN,
        "svensk": Date.CAL_SWEDISH,
        "s": Date.CAL_SWEDISH,
    }

    quality_to_int = {
        "uppskattat": Date.QUAL_ESTIMATED,
        "uppskattad": Date.QUAL_ESTIMATED,
        "bedömt": Date.QUAL_ESTIMATED,
        "bedömd": Date.QUAL_ESTIMATED,
        "beräknat": Date.QUAL_CALCULATED,
        "beräknad": Date.QUAL_CALCULATED,
    }

    def init_strings(self):
        """Define, in Swedish, span and range regular expressions"""
        DateParser.init_strings(self)
        self._numeric = re.compile(r"((\d+)/)?\s*((\d+)/)?\s*(\d+)[/ ]?$")
        # this next RE has the (possibly-slashed) year at the string's start
        self._text2 = re.compile(
            r"((\d+)(/\d+)?)?\s+?%s\s*(\d+)?\s*$" % self._mon_str, re.IGNORECASE
        )
        self._span = re.compile(
            r"(från)?\s*(?P<start>.+)\s*(till|--|–)\s*(?P<stop>.+)", re.IGNORECASE
        )
        self._range = re.compile(
            r"(mellan)\s+(?P<start>.+)\s+och\s+(?P<stop>.+)", re.IGNORECASE
        )


# -------------------------------------------------------------------------
#
# Swedish display class
#
# -------------------------------------------------------------------------
class DateDisplaySv(DateDisplay):
    """
    Swedish language date display class.
    """

    _bce_str = "%s f Kr"

    formats = (
        "ÅÅÅÅ-MM-DD (ISO)",
        "År/mån/dag",
        "År månad dag",
        "År mån dag",
    )
    # this definition must agree with its "_display_calendar" method

    def _display_calendar(self, date_val, long_months, short_months=None, inflect=""):
        # this must agree with its locale-specific "formats" definition

        if short_months is None:
            # Let the short formats work the same as long formats
            short_months = long_months

        if self.format == 0:
            return self.display_iso(date_val)
        elif self.format == 1:
            # numerical: year/month/day (with slashes)
            value = self.dd_dformat01(date_val)
        elif self.format == 2:
            # year month_name day
            value = self.dd_dformat02_sv(date_val, long_months)
        # elif self.format == 3:
        else:
            # year month_abbreviation day
            value = self.dd_dformat03_sv(date_val, short_months)
        if date_val[2] < 0:
            # TODO fix BUG 7064: non-Gregorian calendars wrongly use BCE notation for negative dates
            return self._bce_str % value
        else:
            return value

    def dd_dformat02_sv(self, date_val, long_months):
        # year month_name day
        year = self._slash_year(date_val[2], date_val[3])
        if date_val[0] == 0:
            if date_val[1] == 0:
                return year
            else:
                return "%s %s" % (year, long_months[date_val[1]])
        elif date_val[1] == 0:  # month is zero but day is not (see 8477)
            return self.display_iso(date_val)
        else:
            return "%s %s %s" % (year, long_months[date_val[1]], date_val[0])

    def dd_dformat03_sv(self, date_val, short_months):
        # year month_abbreviation day
        year = self._slash_year(date_val[2], date_val[3])
        if date_val[0] == 0:
            if date_val[1] == 0:
                return year
            else:
                return "%s %s" % (year, short_months[date_val[1]])
        elif date_val[1] == 0:  # month is zero but day is not (see 8477)
            return self.display_iso(date_val)
        else:
            return "%s %s %s" % (year, short_months[date_val[1]], date_val[0])

    display = DateDisplay.display_formatted


# -------------------------------------------------------------------------
#
# Register classes
#
# -------------------------------------------------------------------------
register_datehandler(
    ("sv_SE", "sv_SE.UTF-8", "sv", "Swedish", ("%Y-%m-%d",)),
    DateParserSv,
    DateDisplaySv,
)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
# Copyright (C) 2007-2009  Brian G. Matherly
# Copyright (C) 2009       Gary Burton
# Copyright (C) 2010       Peter Landgren
# Copyright (C) 2011       Adam Stein <adam@csh.rit.edu>
# Copyright (C) 2017       Paul Franklin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# ------------------------------------------------------------------------
#
# Python modules
#
# ------------------------------------------------------------------------
import logging

# ------------------------------------------------------------------------
#
# Gramps modules
#
# ------------------------------------------------------------------------
from gramps.gen.plug.docgen import (
    BaseDoc,
    TextDoc,
    FONT_SERIF,
    PARA_ALIGN_RIGHT,
    PARA_ALIGN_CENTER,
    PARA_ALIGN_JUSTIFY,
    URL_PATTERN,
)
from gramps.gen.utils.image import image_size, image_actual_size, resize_to_jpeg_buffer
from gramps.gen.errors import ReportError
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

LOG = logging.getLogger(".rtfdoc")

# ------------------------------------------------------------------------
#
# Set up to make links clickable
#
# ------------------------------------------------------------------------
_CLICKABLE = r"""{\\field{\\*\\fldinst HYPERLINK "\1"}{\\fldrslt \1}}"""


# ------------------------------------------------------------------------
#
# Functions
#
# ------------------------------------------------------------------------
def twips(length_cm):
    """
    RTF uses a unit called "twips" for its measurements. According to the
    RTF specification, 1 point is 20 twips. This routines converts
    centimeters to twips

    2.54 cm/inch 72pts/inch, 20twips/pt
    """
    return int(((length_cm / 2.54) * 72) + 0.5) * 20


# ------------------------------------------------------------------------
#
# Rich Text Format Document interface. The current interface does not
# use style sheets. Instead it writes raw formatting.
#
# ------------------------------------------------------------------------
class RTFDoc(BaseDoc, TextDoc):
    """
    Opens the file, and writes the header. Builds the color and font tables.
    Fonts are chosen using the MS TrueType fonts, since it is assumed that if
    you are generating RTF, you are probably targeting Word.  This generator
    assumes a Western Europe character set.
    """

    def open(self, filename):
        if filename[-4:] != ".rtf":
            self.filename = filename + ".rtf"
        else:
            self.filename = filename

        try:
            self.file = open(self.filename, "w")
        except IOError as msg:
            errmsg = "%s\n%s" % (_("Could not create %s") % self.filename, msg)
            raise ReportError(errmsg)
        except:
            raise ReportError(_("Could not create %s") % self.filename)

        style_sheet = self.get_style_sheet()

        self.file.write(
            "{\\rtf1\\ansi\\ansicpg1252\\deff0\n"
            "{\\fonttbl\n"
            "{\\f0\\froman\\fcharset0\\fprq0 Times New Roman;}\n"
            "{\\f1\\fswiss\\fcharset0\\fprq0 Arial;}}\n"
            "{\\colortbl\n"
        )

        self.color_map = {}
        index = 1
        self.color_map[(0, 0, 0)] = 0
        self.file.write("\\red0\\green0\\blue0;")
        for style_name in style_sheet.get_paragraph_style_names():
            style = style_sheet.get_paragraph_style(style_name)
            fgcolor = style.get_font().get_color()
            bgcolor = style.get_background_color()
            if fgcolor not in self.color_map:
                self.color_map[fgcolor] = index
                self.file.write("\\red%d\\green%d\\blue%d;" % fgcolor)
                index += 1
            if bgcolor not in self.color_map:
                self.file.write("\\red%d\\green%d\\blue%d;" % bgcolor)
                self.color_map[bgcolor] = index
                index += 1
        self.file.write("}\n")
        self.file.write(
            "\\kerning0\\cf0\\viewkind1"
            + "\\paperw%d" % twips(self.paper.get_size().get_width())
            + "\\paperh%d" % twips(self.paper.get_size().get_height())
            + "\\margl%d" % twips(self.paper.get_left_margin())
            + "\\margr%d" % twips(self.paper.get_right_margin())
            + "\\margt%d" % twips(self.paper.get_top_margin())
            + "\\margb%d" % twips(self.paper.get_bottom_margin())
            + "\\widowctl\n"
        )
        self.in_table = 0
        self.text = ""
        self.p_wide = float(self.paper.get_usable_width())

    # --------------------------------------------------------------------
    #
    # Write the closing brace, and close the file.
    #
    # --------------------------------------------------------------------
    def close(self):
        self.file.write("}\n")
        self.file.close()

    # --------------------------------------------------------------------
    #
    # Force a section page break
    #
    # --------------------------------------------------------------------
    def end_page(self):
        self.file.write("\\sbkpage\n")

    # --------------------------------------------------------------------
    #
    # Starts a paragraph. Instead of using a style sheet, generate
    # the style for each paragraph on the fly. Not the ideal, but it
    # does work.
    #
    # --------------------------------------------------------------------
    def start_paragraph(self, style_name, leader=None):
        self.opened = 0
        style_sheet = self.get_style_sheet()
        para = style_sheet.get_paragraph_style(style_name)

        # build font information

        font = para.get_font()
        size = font.get_size() * 2
        bgindex = self.color_map[para.get_background_color()]
        fgindex = self.color_map[font.get_color()]
        if font.get_type_face() == FONT_SERIF:
            self.font_type = "\\f0"
        else:
            self.font_type = "\\f1"
        self.font_type += "\\fs%d\\cf%d\\cb%d" % (size, fgindex, bgindex)

        if font.get_bold():
            self.font_type += "\\b"
        if font.get_underline():
            self.font_type += "\\ul"
        if font.get_italic():
            self.font_type += "\\i"

        # build paragraph information

        if not self.in_table:
            para_type = "\n\\pard"
        else:
            para_type = "\n\\pard\\intbl"
        if para.get_alignment() == PARA_ALIGN_RIGHT:
            para_type += "\\qr"
        elif para.get_alignment() == PARA_ALIGN_CENTER:
            para_type += "\\qc"
        para_type += (
            "\\ri%d" % twips(para.get_right_margin())
            + "\\li%d" % twips(para.get_left_margin())
            + "\\fi%d" % twips(para.get_first_indent())
        )

        if para.get_alignment() == PARA_ALIGN_JUSTIFY:
            para_type += "\\qj"
        if para.get_padding():
            para_type += "\\sa%d" % twips(para.get_padding() / 2.0)
        if para.get_top_border():
            para_type += "\\brdrt\\brdrs"
        if para.get_bottom_border():
            para_type += "\\brdrb\\brdrs"
        if para.get_left_border():
            para_type += "\\brdrl\\brdrs"
        if para.get_right_border():
            para_type += "\\brdrr\\brdrs"
        if para.get_first_indent():
            para_type += "\\fi%d" % twips(para.get_first_indent())
        if para.get_left_margin():
            para_type += "\\li%d" % twips(para.get_left_margin())
        if para.get_right_margin():
            para_type += "\\ri%d" % twips(para.get_right_margin())

        # the deferred newline while in table cell
        if self.in_table and self.need_nl:
            self.text += "\n\\line}"
            self.need_nl = False

        self.text += para_type

        if leader:
            self.opened = 1
            self.text += "\\tx%d" % twips(para.get_left_margin())
            self.text += "{%s " % self.font_type
            self.text += leader
            # self.write_text(leader)
            # self.file.write(self.text)
            # self.text = ""
            self.text += "\\tab}"
            self.opened = 0

    # --------------------------------------------------------------------
    #
    # Ends a paragraph. Care has to be taken to make sure that the
    # braces are closed properly. The self.opened flag is used to indicate
    # if braces are currently open. If the last write was the end of
    # a bold-faced phrase, braces may already be closed.
    #
    # --------------------------------------------------------------------
    def end_paragraph(self):
        # To deal with newlines in the table cell, we defer them till the
        # beginning of the next paragraph.  This is so that we don't get
        # an excessive space at the end of a table.
        if not self.in_table:
            self.file.write(self.text)
            LOG.debug(
                "end_paragraph: opened: %d write: %s",
                self.opened,
                self.text + "}" if self.opened else "" + "newline",
            )
            if self.opened:
                self.file.write("}")
                self.opened = 0
            self.file.write("\n\\par")
            self.text = ""
        else:
            if self.text == "":
                self.write_text(" ")
            self.need_nl = True

    # --------------------------------------------------------------------
    #
    # Inserts a manual page break
    #
    # --------------------------------------------------------------------
    def page_break(self):
        self.file.write("\\page\n")

    # --------------------------------------------------------------------
    #
    # Starts boldfaced text, enclosed the braces
    #
    # --------------------------------------------------------------------
    def start_bold(self):
        LOG.debug(
            "start_bold: opened: %d saved text: %s",
            self.opened,
            "}" if self.opened else "" + "{%s\\b " % self.font_type,
        )
        if self.opened:
            self.text += "}"
        self.text += "{%s\\b " % self.font_type
        self.opened = 1

    # --------------------------------------------------------------------
    #
    # Ends boldfaced text, closing the braces
    #
    # --------------------------------------------------------------------
    def end_bold(self):
        LOG.debug("end_bold: opened: %d saved text: %s", self.opened, self.text + "}")
        if not self.opened == 1:
            print(self.opened)
            raise RuntimeError
        self.opened = 0
        self.text += "}"

    def start_superscript(self):
        self.text += "{{\\*\\updnprop5801}\\up10 "

    def end_superscript(self):
        self.text += "}"

    # --------------------------------------------------------------------
    #
    # Start a table. Grab the table style, and store it. Keep a flag to
    # indicate that we are in a table. This helps us deal with paragraphs
    # internal to a table. RTF does not require anything to start a
    # table, since a table is treated as a bunch of rows.
    #
    # --------------------------------------------------------------------
    def start_table(self, name, style_name):
        self.in_table = 1
        styles = self.get_style_sheet()
        table_style = styles.get_table_style(style_name)
        self._cells = []
        if self.get_rtl_doc():
            cell_percent = 100.0
        else:
            cell_percent = 0.0
        for cell in range(table_style.get_columns()):
            if self.get_rtl_doc():
                cell_percent -= float(table_style.get_column_width(cell))
            else:
                cell_percent += float(table_style.get_column_width(cell))
            self._cells.append(twips((self.p_wide * cell_percent) / 100.0))
        if self._cells[-1:] == [0]:
            self._cells[-1:] = [twips(self.p_wide)]  # left edge => right edge

    # --------------------------------------------------------------------
    #
    # End a table. Turn off the table flag
    #
    # --------------------------------------------------------------------
    def end_table(self):
        self.in_table = 0

    # --------------------------------------------------------------------
    #
    # Start a row. RTF uses the \trowd to start a row. RTF also specifies
    # all the cell data after it has specified the cell definitions for
    # the row. Therefore it is necessary to keep a list of cell contents
    # that is to be written after all the cells are defined.  It is also
    # necessary to keep a list of the row columns definitions, since for
    # RTL columns the columns which are processed last (in LTR) come first,
    # and we don't know which will be skipped with a non-one "span" argument
    # until we process them
    #
    # --------------------------------------------------------------------
    def start_row(self):
        self.contents = []
        self.columns = []
        self.cell = 0
        self.text = ""
        self.file.write("\\trowd\n")

    # --------------------------------------------------------------------
    #
    # End a row. Write the cell contents, separated by the \cell marker,
    # then terminate the row with a \row marker
    #
    # --------------------------------------------------------------------
    def end_row(self):
        for column in sorted(self.columns):
            for line in column[1]:
                self.file.write(line)
        if self.get_rtl_doc():
            self.contents.reverse()
        self.file.write("{")
        for line in self.contents:
            self.file.write(line)
            self.file.write("\\cell ")
        self.file.write("}\\row\n")

    # --------------------------------------------------------------------
    #
    # Start a cell. Dump out the cell specifics, such as borders. Cell
    # widths are kind of interesting. RTF doesn't specify how wide a cell
    # is, but rather where its right edge is in relationship to the
    # left margin. This means that each cell is the cumlative of the
    # previous cells plus its own width.
    #
    # --------------------------------------------------------------------
    def start_cell(self, style_name, span=1):
        cell_data = []
        styles = self.get_style_sheet()
        s = styles.get_cell_style(style_name)
        if s.get_top_border():
            cell_data.append("\\clbrdrt\\brdrs\\brdrw10\n")
        if s.get_bottom_border():
            cell_data.append("\\clbrdrb\\brdrs\\brdrw10\n")
        if s.get_left_border():
            cell_data.append("\\clbrdrl\\brdrs\\brdrw10\n")
        if s.get_right_border():
            cell_data.append("\\clbrdrr\\brdrs\\brdrw10\n")
        self.need_nl = False
        cell_data.append("\\cellx%d\n" % self._cells[self.cell + span - 1])
        self.columns.append([self._cells[self.cell + span - 1], cell_data])
        self.cell += 1

    # --------------------------------------------------------------------
    #
    # End a cell. Save the current text in the content lists, since data
    # must be saved until all cells are defined.
    #
    # --------------------------------------------------------------------
    def end_cell(self):
        if self.opened == 1:  # We defered paragraph end and now need it.
            self.text += "}"
            self.opened = 0
        self.contents.append(self.text)
        self.text = ""

    # --------------------------------------------------------------------
    #
    # Add a photo. Embed the photo in the document. Use the Python
    # imaging library to load and scale the photo. The image is converted
    # to JPEG, since it is smaller, and supported by RTF. The data is
    # dumped as a string of HEX numbers.
    #
    # --------------------------------------------------------------------
    def add_media(self, name, pos, x_cm, y_cm, alt="", style_name=None, crop=None):
        width, height = image_size(name)

        if (width, height) == (0, 0):
            return

        (act_width, act_height) = image_actual_size(x_cm, y_cm, width, height)

        act_width = twips(act_width)
        act_height = twips(act_height)

        size = [act_width, act_height]
        buf = resize_to_jpeg_buffer(name, size, crop=crop)
        # The size may change because of cropping or keeping the ratio
        act_width = size[0]
        act_height = size[1]

        if self.in_table:
            self.text += "{\\*\\shppict{\\pict\\jpegblip"
            self.text += "\\picwgoal%d\\pichgoal%d\n" % (act_width, act_height)
            index = 1
            for i in buf:
                self.text += "%02x" % i
                if index % 32 == 0:
                    self.text += "\n"
                index = index + 1
            self.text += "}}\\par\n"
        else:
            self.file.write("{\\*\\shppict{\\pict\\jpegblip")
            self.file.write("\\picwgoal%d\\pichgoal%d\n" % (act_width, act_height))
            index = 1
            for i in buf:
                self.file.write("%02x" % i)
                if index % 32 == 0:
                    self.file.write("\n")
                index = index + 1
            self.file.write("}}\\par\n")

        if len(alt):
            self.file.write("%s\n\\par\n" % "\\par".join(alt))

    def write_styled_note(
        self, styledtext, format, style_name, contains_html=False, links=False
    ):
        """
        Convenience function to write a styledtext to the RTF doc.
        styledtext : assumed a StyledText object to write
        format : = 0 : Flowed, = 1 : Preformatted
        style_name : name of the style to use for default presentation
        contains_html: bool, the backend should not check if html is present.
            If contains_html=True, then the textdoc is free to handle that in
            some way. Eg, a textdoc could remove all tags, or could make sure
            a link is clickable. RTFDoc prints the html without handling it
        links: bool, make URLs clickable if True
        """
        text = str(styledtext)
        self.start_paragraph(style_name)
        linenb = 1
        for line in text.split("\n"):
            [line, sigcount] = process_spaces(line, format)
            if sigcount == 0:
                if self.in_table:
                    #    # Add LF when in table as in indiv_complete report
                    self.write_text("\n")
                self.end_paragraph()
                self.start_paragraph(style_name)
                linenb = 1
            else:
                if linenb > 1:
                    self.write_text("\n")
                self.write_text(line, links=links)
                linenb += 1
        # FIXME: I don't understand why these newlines are necessary.
        # It may be related to the behaviour of end_paragraph inside tables,
        # and write_text converting \n to end paragraph.
        # This code prevents the whole document going wrong, but seems to
        # produce an extra paragraph mark at the end of each table cell.
        if self.in_table:
            #    # Add LF when in table as in indiv_complete report
            self.write_text("\n")
        self.end_paragraph()

    # --------------------------------------------------------------------
    #
    # Writes text. If braces are not currently open, open them. Loop
    # character by character (terribly inefficient, but it works). If a
    # character is 8 bit (>127), convert it to a hex representation in
    # the form of \`XX. Make sure to escape braces.
    #
    # --------------------------------------------------------------------
    def write_text(self, text, mark=None, links=False):
        # Convert to unicode, just in case it's not. Fix of bug 2449.
        text = str(text)
        LOG.debug("write_text: opened: %d input text: %s", self.opened, text)
        if self.opened == 0:
            self.opened = 1
            self.text += "{%s " % self.font_type

        for i in text:
            if ord(i) > 127:
                if ord(i) < 256:
                    self.text += "\\'%2x" % ord(i)
                else:
                    # If (uni)code with more than 8 bits:
                    # RTF req valus in decimal, not hex.
                    self.text += "{\\uc0\\u%d}" % ord(i)
            elif i == "\n":
                self.text += "\n\\par "
            elif i == "{" or i == "}" or i == "\\":
                self.text += "\\%s" % i
            else:
                self.text += i

        if links is True:
            import re

            self.text = re.sub(URL_PATTERN, _CLICKABLE, self.text)
        LOG.debug("write_text, exit: opened: %d saved text: %s", self.opened, self.text)


def process_spaces(line, format):
    """
    Function to process spaces in text lines for flowed and pre-formatted notes.
    line : text to process
    format : = 0 : Flowed, = 1 : Preformatted

    If the text is flowed (format==0), then leading spaces
    are removed, and multiple spaces are reduced to one.
    If the text is pre-formatted (format==1). then all spaces are preserved

    Note that xml is just treated like any other text, because it will be from
    the original note, and it is just printed, not interpreted.
    Returns the processed text, and the number of significant
    (i.e. non-white-space) chars.
    """
    txt = ""
    xml = False
    space = False
    sigcount = 0
    # we loop through every character, which is very inefficient, but an
    # attempt to use a regex replace didn't always work.
    for char in line:
        if char == " " or char == "\t":
            if format == 1:
                txt += char
            elif format == 0 and sigcount == 0:
                pass
            elif format == 0 and space is False:
                space = True
                txt += char
            elif format == 0 and space is True:
                pass
        else:
            sigcount += 1
            space = False
            txt += char
    return [txt, sigcount]



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""unittest for grampstype"""

from __future__ import annotations
import unittest

from ..grampstype import GrampsType

# some simple map items to test with
vals = "zz ab cd ef".split()
keys = list(range(len(vals)))
MAP = [(k, v * 2, v) for (k, v) in zip(keys, vals)]
BLIST = [1, 3]


class GT0(GrampsType):
    _DEFAULT = 1  # just avoiding the pre-coded 0
    _CUSTOM = 3  # just avoiding the pre-coded 0
    _DATAMAP = MAP


# NOTE: this type of code might be used in a migration utility
#   to allow conversions or other handling of retired type-values
# A migration utility might instantiate several of these with
#   varying blacklist-specs
class GT1(GT0):
    _BLACKLIST: list[int] | None = BLIST


class GT2(GT1):
    _BLACKLIST = None


class Test1(unittest.TestCase):
    # some basic tests
    def test_basic(self):
        self.gt = GT0()
        self.assertTrue(isinstance(self.gt, GrampsType))
        # spot-check that MAPs get built
        e = len(keys)
        g = len(self.gt._E2IMAP)
        self.assertEqual(g, e)

    # init sets values for int, str, tuple
    # (we ignore instance here -- maybe SB tested, too?)
    # this test depends on having _DEFAULT=1, _CUSTOM=3
    # NB: tuple tests w/ lengths < 2 fail before release 10403
    def test_init_value(self):
        for i, v, u in (
            (None, 1, "abab"),  # all DEFAULT
            (0, 0, "zzzz"),
            (1, 1, "abab"),
            ("efef", 3, "efef"),  # matches CUSTOM
            ("zzzz", 0, "zzzz"),
            ("x", 3, "x"),  # nomatch gives CUSTOM
            ("", 3, ""),  # nomatch gives CUSTOM
            ((0, "zero"), 0, "zzzz"),  # normal behavior
            ((2,), 2, "cdcd"),  # DEFAULT-string, just like int
            ((), 1, "abab"),  # DEFAULT-pair
        ):
            self.gt = GT0(i)
            g = self.gt.value
            self.assertEqual(g, v)
            g = self.gt.string
            self.assertEqual(g, u)


# test blacklist functionality added to enable fix of bug #1680
class Test2(unittest.TestCase):
    def test_blacklist(self):
        self.gt = GT1()
        # check that MAPs have lengths reduced by blacklist
        e = len(keys) - len(BLIST)
        g = len(self.gt._E2IMAP)
        self.assertEqual(g, e)

        self.ub = GT2()
        # check that these MAPS are now un-blacklisted
        e = len(keys)
        g = len(self.ub._E2IMAP)
        self.assertEqual(g, e)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2012       Doug Blank <doug.blank@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Find the latest git revision.
"""

import subprocess


def get_git_revision(path=""):
    """
    Return the short commit hash of the latest commit.
    """
    stdout = ""
    command = ["git", "log", "-1", "--format=%h", path]
    try:
        proc = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        (stdout, stderr) = proc.communicate()
    except OSError:
        return ""  # subprocess failed
    # subprocess worked
    if stdout and len(stdout) > 0:  # has output
        try:
            stdout = stdout.decode("utf-8", errors="replace")
        except UnicodeDecodeError:
            pass
        return "-" + stdout if stdout else ""
    else:  # no output from git log
        return ""



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2015        Nick Hall
# Copyright (C) 2017-       Serge Noiraud
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

__all__ = ["PlaceWithin"]

# -------------------------------------------------------------------------
#
# Standard python modules
#
# -------------------------------------------------------------------------
import logging

_LOG = logging.getLogger(".widgets.placewithin")

# -------------------------------------------------------------------------
#
# GTK/Gnome modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..selectors import SelectorFactory
from gramps.gen.display.place import displayer as _pd
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext


# -------------------------------------------------------------------------
#
# PlaceWithin class
#
# -------------------------------------------------------------------------
class PlaceWithin(Gtk.Box):
    def __init__(self, dbstate, uistate, track):
        Gtk.Box.__init__(self)
        self.dbstate = dbstate
        self.uistate = uistate
        self.track = track
        self.last = ""
        # initial tooltip when no place already selected.
        self.tooltip = _(
            "Matches places within a given distance"
            " of the active place. You have no active place."
        )
        self.set_tooltip_text(self.tooltip)
        self.entry = Gtk.Entry()
        self.entry.set_max_length(3)
        self.entry.set_width_chars(5)
        self.entry.connect("changed", self.entry_change)
        self.pack_start(self.entry, True, True, 0)
        self.unit = Gtk.ComboBoxText()
        list(map(self.unit.append_text, [_("kilometers"), _("miles"), _("degrees")]))
        self.unit.set_active(0)
        self.pack_start(self.unit, False, True, 0)
        self.show_all()

    def get_value(self):
        value = self.entry.get_text()
        if value == "":
            value = "0"
        return int(value), self.unit.get_active()

    def set_value(self, value, unit):
        self.entry.set_text(str(value))
        self.unit.set_active(int(unit))

    def entry_change(self, entry):
        value = entry.get_text()
        if value.isnumeric() or value == "":
            self.last = value  # This entry is numeric and valid.
        else:
            entry.set_text(self.last)  # reset to the last valid entry
        _db = self.dbstate.db
        active_reference = self.uistate.get_active("Place")
        place_name = None
        if active_reference:
            place = _db.get_place_from_handle(active_reference)
            place_name = _pd.display(self.dbstate.db, place)
        if place_name is None:
            self.set_tooltip_text(self.tooltip)
        else:
            self.set_tooltip_text(place_name)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Custom Filter Editor tool.
"""

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------

# ------------------------------------------------------------------------
#
# Set up logging
#
# ------------------------------------------------------------------------
import logging

log = logging.getLogger(".filtereditor")

# -------------------------------------------------------------------------
#
# GTK/GNOME
#
# -------------------------------------------------------------------------
from gi.repository import Gtk
from gi.repository import Gdk
from gi.repository import GObject

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext
from gramps.gen.filters import GenericFilterFactory, FilterList, reload_custom_filters
from gramps.gen.filters.rules._matchesfilterbase import MatchesFilterBase
from ..listmodel import ListModel
from ..managedwindow import ManagedWindow
from ..dialog import QuestionDialog, ErrorDialog
from gramps.gen.const import RULE_GLADE, URL_MANUAL_PAGE
from ..display import display_help
from gramps.gen.errors import WindowActiveError, FilterError
from gramps.gen.lib import (
    AttributeType,
    EventType,
    FamilyRelType,
    NameOriginType,
    NameType,
    NoteType,
    PlaceType,
)
from gramps.gen.filters import rules
from ..autocomp import StandardCustomSelector, fill_entry
from ..selectors import SelectorFactory
from gramps.gen.display.name import displayer as _nd
from gramps.gen.display.place import displayer as _pd
from gramps.gen.utils.db import family_name
from gramps.gen.utils.string import conf_strings
from ..widgets import DateEntry
from gramps.gen.datehandler import displayer
from gramps.gen.config import config
from gramps.gui.widgets.persistenttreeview import PersistentTreeView

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------
WIKI_HELP_PAGE = "%s_-_Filters" % URL_MANUAL_PAGE
WIKI_HELP_SEC = _("Add_Rule_dialog", "manual")
WIKI_HELP_SEC2 = _("Define_Filter_dialog", "manual")
WIKI_HELP_SEC3 = _("Custom_Filters", "manual")

# dictionary mapping FILTER_TYPE of views to Filter window name
_TITLES = {
    "Person": _("Person Filters"),
    "Family": _("Family Filters"),
    "Event": _("Event Filters"),
    "Place": _("Place Filters"),
    "Source": _("Source Filters"),
    "Media": _("Media Filters"),
    "Repository": _("Repository Filters"),
    "Note": _("Note Filters"),
    "Citation": _("Citation Filters"),
}

_name2typeclass = {
    _("Personal event:"): EventType,
    _("Family event:"): EventType,
    _("Event type:"): EventType,
    _("Personal attribute:"): AttributeType,
    _("Family attribute:"): AttributeType,
    _("Event attribute:"): AttributeType,
    _("Media attribute:"): AttributeType,
    _("Relationship type:"): FamilyRelType,
    _("Note type:"): NoteType,
    _("Name type:"): NameType,
    _("Surname origin type:"): NameOriginType,
    _("Place type:"): PlaceType,
}


# -------------------------------------------------------------------------
#
# MyBoolean - check button with standard interface
#
# -------------------------------------------------------------------------
class MyBoolean(Gtk.CheckButton):
    def __init__(self, label=None):
        Gtk.CheckButton.__init__(self)
        self.set_label(label)
        self.show()

    def get_text(self):
        """
        Return the text to save.

        It should be the same no matter the present locale (English or numeric
        types).
        This class sets this to get_display_text, but when localization
        is an issue (events/attr/etc types) then it has to be overridden.

        """
        return str(int(self.get_active()))

    def set_text(self, val):
        """
        Set the selector state to display the passed value.
        """
        is_active = bool(int(val))
        self.set_active(is_active)


# -------------------------------------------------------------------------
#
# MyInteger - spin button with standard interface
#
# -------------------------------------------------------------------------
class MyInteger(Gtk.SpinButton):
    def __init__(self, min, max):
        Gtk.SpinButton.__init__(self)
        self.set_adjustment(
            Gtk.Adjustment(value=min, lower=min, upper=max, step_increment=1)
        )
        self.show()

    def get_text(self):
        return str(self.get_value_as_int())

    def set_text(self, val):
        self.set_value(int(val))


# -------------------------------------------------------------------------
#
# MyFilters - Combo box with list of filters with a standard interface
#
# -------------------------------------------------------------------------
class MyFilters(Gtk.ComboBox):
    """
    Class to present a combobox of selectable filters.
    """

    def __init__(self, filters, filter_name=None):
        """
        Construct the combobox from the entries of the filters list.

        Filter_name is name of calling filter.
        If filter_name is given, it will be excluded from the dropdown box.

        """
        Gtk.ComboBox.__init__(self)
        store = Gtk.ListStore(GObject.TYPE_STRING)
        self.set_model(store)
        cell = Gtk.CellRendererText()
        self.pack_start(cell, True)
        self.add_attribute(cell, "text", 0)
        # remove own name from the list if given.
        self.flist = sorted(
            f.get_name()
            for f in filters
            if filter_name is None or f.get_name() != filter_name
        )

        for fname in self.flist:
            store.append(row=[fname])
        self.set_active(0)
        self.show()

    def get_text(self):
        active = self.get_active()
        if active < 0:
            return ""
        return self.flist[active]

    def set_text(self, val):
        if val in self.flist:
            self.set_active(self.flist.index(val))


# -------------------------------------------------------------------------
#
# MyList - Combo box to allow entries
#
# -------------------------------------------------------------------------
class MyList(Gtk.ComboBox):
    def __init__(self, clist, clist_trans, default=0):
        Gtk.ComboBox.__init__(self)
        store = Gtk.ListStore(GObject.TYPE_STRING)
        self.set_model(store)
        cell = Gtk.CellRendererText()
        self.pack_start(cell, True)
        self.add_attribute(cell, "text", 0)
        self.clist = clist
        for name in clist_trans:
            store.append(row=[name])
        self.set_active(default)
        self.show()

    def get_text(self):
        active = self.get_active()
        return self.clist[active]

    def set_text(self, val):
        if val in self.clist:
            self.set_active(self.clist.index(val))


# -------------------------------------------------------------------------
#
# MyLesserEqualGreater - Combo box to allow selection of "<", "=", or ">"
#
# -------------------------------------------------------------------------
class MyLesserEqualGreater(Gtk.ComboBox):
    def __init__(self, default=0):
        Gtk.ComboBox.__init__(self)
        store = Gtk.ListStore(GObject.TYPE_STRING)
        self.set_model(store)
        cell = Gtk.CellRendererText()
        self.pack_start(cell, True)
        self.add_attribute(cell, "text", 0)
        self.clist = ["less than", "equal to", "greater than"]
        self.clist_trans = [_("less than"), _("equal to"), _("greater than")]
        for name in self.clist_trans:
            store.append(row=[name])
        self.set_active(default)
        self.show()

    def get_text(self):
        active = self.get_active()
        if active < 0:
            return "equal to"
        return self.clist[active]

    def set_text(self, val):
        if val in self.clist:
            self.set_active(self.clist.index(val))
        else:
            self.set_active(self.clist.index("equal to"))


# -------------------------------------------------------------------------
#
# MyPlaces - AutoCombo text entry with list of places attached. Provides
#            a standard interface
#
# -------------------------------------------------------------------------
class MyPlaces(Gtk.Entry):
    def __init__(self, places):
        Gtk.Entry.__init__(self)

        fill_entry(self, places)
        self.show()


# -------------------------------------------------------------------------
#
# MyID - Person/GRAMPS ID selection box with a standard interface
#
# -------------------------------------------------------------------------
class MyID(Gtk.Box):
    _invalid_id_txt = _("Not a valid ID")
    _empty_id_txt = _invalid_id_txt

    obj_name = {
        "Person": _("Person"),
        "Family": _("Family"),
        "Event": _("Event"),
        "Place": _("Place"),
        "Source": _("Source"),
        "Media": _("Media"),
        "Repository": _("Repository"),
        "Note": _("Note"),
        "Citation": _("Citation"),
    }

    def __init__(self, dbstate, uistate, track, namespace="Person"):
        Gtk.Box.__init__(self)
        self.set_orientation(Gtk.Orientation.HORIZONTAL)
        self.set_homogeneous(False)
        self.set_spacing(6)
        self.dbstate = dbstate
        self.db = dbstate.db
        self.uistate = uistate
        self.track = track

        self.namespace = namespace
        self.entry = Gtk.Entry()
        self.entry.show()
        self.button = Gtk.Button()
        self.button.set_label(_("Select..."))
        self.button.connect("clicked", self.button_press)
        self.button.show()
        self.pack_start(self.entry, True, True, 0)
        self.add(self.button)
        self.button.set_tooltip_text(
            _("Select %s from a list") % self.obj_name[namespace]
        )
        self.show()
        self.set_text("")

    def button_press(self, obj):
        obj_class = self.namespace
        selector = SelectorFactory(obj_class)
        inst = selector(self.dbstate, self.uistate, self.track)
        val = inst.run()
        if val is None:
            self.set_text("")
        else:
            self.set_text(val.get_gramps_id())

    def get_text(self):
        return str(self.entry.get_text())

    def name_from_gramps_id(self, gramps_id):
        if self.namespace == "Person":
            person = self.db.get_person_from_gramps_id(gramps_id)
            name = _nd.display_name(person.get_primary_name())
        elif self.namespace == "Family":
            family = self.db.get_family_from_gramps_id(gramps_id)
            name = family_name(family, self.db)
        elif self.namespace == "Event":
            event = self.db.get_event_from_gramps_id(gramps_id)
            name = str(event.get_type)
        elif self.namespace == "Place":
            place = self.db.get_place_from_gramps_id(gramps_id)
            name = _pd.display(self.db, place)
        elif self.namespace == "Source":
            source = self.db.get_source_from_gramps_id(gramps_id)
            name = source.get_title()
        elif self.namespace == "Citation":
            citation = self.db.get_citation_from_gramps_id(gramps_id)
            name = citation.get_page()
        elif self.namespace == "Media":
            obj = self.db.get_media_from_gramps_id(gramps_id)
            name = obj.get_path()
        elif self.namespace == "Repository":
            repo = self.db.get_repository_from_gramps_id(gramps_id)
            name = repo.get_name()
        elif self.namespace == "Note":
            note = self.db.get_note_from_gramps_id(gramps_id)
            name = note.get()
        return name

    def set_text(self, val):
        if not val:
            self.entry.set_tooltip_text(self._empty_id_txt)
        else:
            try:
                name = self.name_from_gramps_id(val)
                self.entry.set_tooltip_text(name)
            except AttributeError:
                self.entry.set_tooltip_text(self._invalid_id_txt)
        self.entry.set_text(val)


# -------------------------------------------------------------------------
#
# MySource - select ID of sources with a standard interface
#
# -------------------------------------------------------------------------
class MySource(MyID):
    _empty_id_txt = _(
        "Give or select a source ID, leave empty to find objects" " with no source."
    )

    def __init__(self, dbstate, uistate, track):
        MyID.__init__(self, dbstate, uistate, track, namespace="Source")
        self.entry.set_tooltip_text(self._empty_id_txt)


# -------------------------------------------------------------------------
#
# MySelect
#
# -------------------------------------------------------------------------
class MySelect(Gtk.ComboBox):
    def __init__(self, type_class, additional):
        # we need to inherit and have an combobox with an entry
        Gtk.ComboBox.__init__(self, has_entry=True)
        self.type_class = type_class
        self.sel = StandardCustomSelector(
            type_class._I2SMAP,
            self,
            type_class._CUSTOM,
            type_class._DEFAULT,
            additional,
            type_class._MENU,
        )
        self.show()

    def get_text(self):
        return self.type_class(self.sel.get_values()).xml_str()

    def set_text(self, val):
        tc = self.type_class()
        tc.set_from_xml_str(val)
        self.sel.set_values((int(tc), str(tc)))


# -------------------------------------------------------------------------
#
# MyEntry
#
# -------------------------------------------------------------------------
class MyEntry(Gtk.Entry):
    def __init__(self):
        Gtk.Entry.__init__(self)
        self.show()


# -------------------------------------------------------------------------
#
# EditRule
#
# -------------------------------------------------------------------------
class EditRule(ManagedWindow):
    def __init__(
        self,
        namespace,
        dbstate,
        uistate,
        track,
        filterdb,
        val,
        label,
        update,
        filter_name,
    ):
        ManagedWindow.__init__(self, uistate, track, EditRule)
        self.namespace = namespace
        self.dbstate = dbstate
        self.db = dbstate.db
        self.filterdb = filterdb
        self.update_rule = update
        self.filter_name = filter_name

        self.active_rule = val
        self.define_glade("rule_editor", RULE_GLADE)

        self.set_window(
            self.get_widget("rule_editor"), self.get_widget("rule_editor_title"), label
        )
        self.setup_configs("interface.edit-rule", 600, 450)
        self.window.hide()
        self.valuebox = self.get_widget("valuebox")
        self.rname_filter = self.get_widget("ruletreefilter")
        self.rule_name = self.get_widget("rulename")
        self.description = self.get_widget("description")

        objectlist = self.get_widget("ruletree")
        self.rname = PersistentTreeView(uistate, "filt_edit")
        scrolledwindow = self.get_widget("box2")
        scrolledwindow.remove(objectlist)
        scrolledwindow.add(self.rname)

        self.notebook = Gtk.Notebook()
        self.notebook.set_show_tabs(0)
        self.notebook.set_show_border(0)
        self.notebook.show()
        self.valuebox.pack_start(self.notebook, True, True, 0)
        self.page_num = 0
        self.page = []
        self.class2page = {}

        if self.namespace == "Person":
            class_list = rules.person.editor_rule_list
        elif self.namespace == "Family":
            class_list = rules.family.editor_rule_list
        elif self.namespace == "Event":
            class_list = rules.event.editor_rule_list
        elif self.namespace == "Source":
            class_list = rules.source.editor_rule_list
        elif self.namespace == "Citation":
            class_list = rules.citation.editor_rule_list
        elif self.namespace == "Place":
            class_list = rules.place.editor_rule_list
        elif self.namespace == "Media":
            class_list = rules.media.editor_rule_list
        elif self.namespace == "Repository":
            class_list = rules.repository.editor_rule_list
        elif self.namespace == "Note":
            class_list = rules.note.editor_rule_list

        for class_obj in class_list:
            arglist = class_obj.labels
            vallist = []
            tlist = []
            pos = 0
            l2 = Gtk.Label(label=class_obj.name, halign=Gtk.Align.START)
            l2.show()
            grid = Gtk.Grid()
            grid.set_border_width(12)
            grid.set_column_spacing(6)
            grid.set_row_spacing(6)
            grid.show()
            for v in arglist:
                if isinstance(v, tuple):
                    # allows filter to create its own GUI element
                    l = Gtk.Label(label=v[0], halign=Gtk.Align.END)
                else:
                    l = Gtk.Label(label=v, halign=Gtk.Align.END)
                l.show()
                if v == _("Place:"):
                    t = MyPlaces([])
                elif v in [_("Reference count:"), _("Number of instances:")]:
                    t = MyInteger(0, 999)
                elif v == _("Reference count must be:"):
                    t = MyLesserEqualGreater()
                elif v == _("Number must be:"):
                    t = MyLesserEqualGreater(2)
                elif v == _("Number of generations:"):
                    t = MyInteger(1, 32)
                elif v == _("ID:"):
                    t = MyID(self.dbstate, self.uistate, self.track, self.namespace)
                elif v == _("Source ID:"):
                    t = MySource(self.dbstate, self.uistate, self.track)
                elif v == _("Filter name:"):
                    t = MyFilters(
                        self.filterdb.get_filters(self.namespace), self.filter_name
                    )
                # filters of another namespace, name may be same as caller!
                elif v == _("Person filter name:"):
                    t = MyFilters(self.filterdb.get_filters("Person"))
                elif v == _("Family filter name:"):
                    t = MyFilters(self.filterdb.get_filters("Family"))
                elif v == _("Event filter name:"):
                    t = MyFilters(self.filterdb.get_filters("Event"))
                elif v == _("Source filter name:"):
                    t = MyFilters(self.filterdb.get_filters("Source"))
                elif v == _("Repository filter name:"):
                    t = MyFilters(self.filterdb.get_filters("Repository"))
                elif v == _("Place filter name:"):
                    t = MyFilters(self.filterdb.get_filters("Place"))
                elif v in _name2typeclass:
                    additional = None
                    if v in (
                        _("Event type:"),
                        _("Personal event:"),
                        _("Family event:"),
                    ):
                        additional = self.db.get_event_types()
                    elif v == _("Personal attribute:"):
                        additional = self.db.get_person_attribute_types()
                    elif v == _("Family attribute:"):
                        additional = self.db.get_family_attribute_types()
                    elif v == _("Event attribute:"):
                        additional = self.db.get_event_attribute_types()
                    elif v == _("Media attribute:"):
                        additional = self.db.get_media_attribute_types()
                    elif v == _("Relationship type:"):
                        additional = self.db.get_family_relation_types()
                    elif v == _("Note type:"):
                        additional = self.db.get_note_types()
                    elif v == _("Name type:"):
                        additional = self.db.get_name_types()
                    elif v == _("Surname origin type:"):
                        additional = self.db.get_origin_types()
                    elif v == _("Place type:"):
                        additional = sorted(
                            self.db.get_place_types(), key=lambda s: s.lower()
                        )
                    t = MySelect(_name2typeclass[v], additional)
                elif v == _("Inclusive:"):
                    t = MyBoolean(_("Include selected Gramps ID"))
                elif v == _("Case sensitive:"):
                    t = MyBoolean(_("Use exact case of letters"))
                elif v == _("Regular-Expression matching:"):
                    t = MyBoolean(_("Use regular expression"))
                elif v == _("Include Family events:"):
                    t = MyBoolean(_("Also family events where person is spouse"))
                elif v == _("Primary Role:"):
                    t = MyBoolean(_("Only include primary participants"))
                elif v == _("Tag:"):
                    taglist = [""]
                    taglist = taglist + [
                        tag.get_name() for tag in dbstate.db.iter_tags()
                    ]
                    t = MyList(taglist, taglist)
                elif v == _("Confidence level:"):
                    t = MyList(
                        list(map(str, list(range(5)))),
                        [_(conf_strings[i]) for i in range(5)],
                    )
                elif v == _("Date:"):
                    t = DateEntry(self.uistate, self.track)
                elif v == _("Day of Week:"):
                    long_days = displayer.long_days
                    days_of_week = long_days[2:] + long_days[1:2]
                    t = MyList(list(map(str, range(7))), days_of_week)
                elif v == _("Units:"):
                    t = MyList([0, 1, 2], [_("kilometers"), _("miles"), _("degrees")])
                elif isinstance(v, tuple):
                    # allow filter to create its own GUI element
                    t = v[1](self.db)
                else:
                    t = MyEntry()
                t.set_hexpand(True)
                tlist.append(t)
                grid.attach(l, 0, pos, 1, 1)
                grid.attach(t, 1, pos, 1, 1)
                pos += 1

            use_regex = None
            use_case = None
            if class_obj.allow_regex:
                use_regex = Gtk.CheckButton(label=_("Use regular expressions"))
                tip = _(
                    "Interpret the contents of string fields as regular "
                    "expressions:\n"
                    ".\tA decimal point will match any character.\n"
                    "?\tA question mark will match zero or one occurences "
                    "of the previous character or group.\n"
                    "*\tAn asterisk will match zero or more occurences.\n"
                    "+\tA plus sign will match one or more occurences.\n"
                    "()\tUse parentheses to group expressions.\n"
                    "|\tSpecify alternatives using a vertical bar.\n"
                    "^\tA caret will match the start of a line.\n"
                    "$\tA dollar sign will match the end of a line."
                )
                use_regex.set_tooltip_text(tip)
                grid.attach(use_regex, 1, pos, 1, 1)

                pos += 1
                use_case = Gtk.CheckButton(label=_("Case sensitive"))
                grid.attach(use_case, 1, pos, 1, 1)
                use_regex.connect("toggled", self.regex_selection, use_case)
                use_case.set_sensitive(False)

            self.page.append((class_obj, vallist, tlist, use_regex, use_case))

            # put the grid into a scrollable area:
            scrolled_win = Gtk.ScrolledWindow()
            scrolled_win.add(grid)
            scrolled_win.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
            scrolled_win.show()
            self.notebook.append_page(scrolled_win, Gtk.Label(label=class_obj.name))
            self.class2page[class_obj] = self.page_num
            self.page_num = self.page_num + 1
        self.page_num = 0
        self.store = Gtk.TreeStore(GObject.TYPE_STRING, GObject.TYPE_PYOBJECT)
        self.ruletree_filter = self.store.filter_new()
        self.ruletree_filter.set_visible_func(self.rtree_visible_func)
        self.selection = self.rname.get_selection()
        col = Gtk.TreeViewColumn(_("Rule Name"), Gtk.CellRendererText(), text=0)
        self.rname.append_column(col)
        self.rname.set_model(self.ruletree_filter)

        prev = None
        last_top = None

        top_level = {}
        top_node = {}

        #
        # If editing a rule, get the name so that we can select it later
        #
        sel_node = None
        if self.active_rule is not None:
            self.sel_class = self.active_rule.__class__
        else:
            self.sel_class = None

        keys = sorted(class_list, key=lambda x: x.name, reverse=True)
        catlist = sorted(set(class_obj.category for class_obj in keys))

        for category in catlist:
            top_node[category] = self.store.insert_after(None, last_top)
            top_level[category] = []
            last_top = top_node[category]
            self.store.set(last_top, 0, category, 1, "")

        for class_obj in keys:
            category = class_obj.category
            top_level[category].append(class_obj.name)
            node = self.store.insert_after(top_node[category], prev)
            self.store.set(node, 0, class_obj.name, 1, class_obj)

            # if this is an edit rule, save the node
            if class_obj == self.sel_class:
                sel_node = (top_node[category], node)

        if sel_node:
            self.select_iter(sel_node)
            page = self.class2page[self.active_rule.__class__]
            self.notebook.set_current_page(page)
            self.display_values(self.active_rule.__class__)
            (class_obj, vallist, tlist, use_regex, use_case) = self.page[page]
            r = list(self.active_rule.values())
            for i in range(0, min(len(tlist), len(r))):
                tlist[i].set_text(r[i])
            if class_obj.allow_regex:
                use_regex.set_active(self.active_rule.use_regex)
                use_case.set_active(self.active_rule.use_case)
                self.regex_selection()

        self.selection.connect("changed", self.on_node_selected)
        self.rname.connect("button-press-event", self._button_press)
        self.rname.connect("key-press-event", self._key_press)
        self.get_widget("rule_editor_ok").connect("clicked", self.rule_ok)
        self.get_widget("rule_editor_cancel").connect("clicked", self.close_window)
        self.get_widget("rule_editor_help").connect("clicked", self.on_help_clicked)
        self.rname_filter.connect("changed", self.on_rname_filter_changed)

        self._set_size()
        config.register("interface.edit-rule-pane", 205)
        panepos = config.get("interface.edit-rule-pane")
        self.get_widget("hpaned1").set_position(panepos)
        self.rname.restore_column_size()
        self.show()

    def regex_selection(self, widget=None, use_case=None):
        if use_case:
            if widget and widget.get_active():
                use_case.set_sensitive(True)
            else:
                use_case.set_active(False)
                use_case.set_sensitive(False)

    def select_iter(self, data):
        """
        Workaround to get self.selection to move to iter row.
        self.selection.select_iter(iter) did not work, so we first
        select the top_node iter, expand it, and then select_iter.
        """
        top_node, iter = data
        self.selection.select_iter(top_node)
        self.expand_collapse()
        self.selection.select_iter(iter)

    def _button_press(self, obj, event):
        if event.type == Gdk.EventType.DOUBLE_BUTTON_PRESS and event.button == 1:
            return self.expand_collapse()

    def _key_press(self, obj, event):
        if event.keyval in (Gdk.KEY_Return, Gdk.KEY_KP_Enter):
            return self.expand_collapse()
        return False

    def expand_collapse(self):
        """
        Expand or collapse the selected parent name node.
        Return True if change done, False otherwise
        """
        store, paths = self.selection.get_selected_rows()
        if paths and len(paths[0].get_indices()) == 1:
            if self.rname.row_expanded(paths[0]):
                self.rname.collapse_row(paths[0])
            else:
                self.rname.expand_row(paths[0], 0)
            return True
        return False

    def on_help_clicked(self, obj):
        """
        Display the relevant portion of Gramps manual.
        """
        display_help(webpage=WIKI_HELP_PAGE, section=WIKI_HELP_SEC)

    def close_window(self, obj):
        panepos = self.get_widget("hpaned1").get_position()
        config.set("interface.edit-rule-pane", panepos)
        self.close()

    def on_node_selected(self, obj):
        """
        Update the informational display on the right hand side of the dialog
        box with the description of the selected report.
        """
        store, node = self.selection.get_selected()
        if node:
            class_obj = store.get_value(node, 1)
            self.display_values(class_obj)

    def on_rname_filter_changed(self, obj):
        """
        Update the ruletree based on the filter.
        """
        self.ruletree_filter.refilter()

    def display_values(self, class_obj):
        if class_obj in self.class2page:
            page = self.class2page[class_obj]
            self.notebook.set_current_page(page)
            self.valuebox.set_sensitive(1)
            self.rule_name.set_text(class_obj.name)
            self.description.set_text(class_obj.description)
        else:
            self.valuebox.set_sensitive(0)
            self.rule_name.set_text(_("No rule selected"))
            self.description.set_text("")

    def rule_ok(self, obj):
        if self.rule_name.get_text() == _("No rule selected"):
            return

        try:
            page = self.notebook.get_current_page()
            (class_obj, vallist, tlist, use_regex, use_case) = self.page[page]
            value_list = [str(sclass.get_text()) for sclass in tlist]
            if class_obj.allow_regex:
                new_rule = class_obj(
                    value_list, use_regex.get_active(), use_case.get_active()
                )
            else:
                new_rule = class_obj(value_list)

            self.update_rule(self.active_rule, new_rule)
            self.close()
        except KeyError:
            pass

    def rtree_visible_func(self, model, iter, data):
        """
        Callback function to determine if a row of the tree is visible
        """
        filter_text = self.rname_filter.get_text()
        tree_text = model[iter][0]
        children = model[iter].iterchildren()
        result = (
            not tree_text or children.iter or filter_text.lower() in tree_text.lower()
        )
        return result


# -------------------------------------------------------------------------
#
# EditFilter
#
# -------------------------------------------------------------------------
class EditFilter(ManagedWindow):
    def __init__(
        self,
        namespace,
        dbstate,
        uistate,
        track,
        gfilter,
        filterdb,
        update=None,
        selection_callback=None,
    ):
        ManagedWindow.__init__(self, uistate, track, self)
        self.namespace = namespace
        self.update = update
        self.dbstate = dbstate
        self.db = dbstate.db
        self.filter = gfilter
        self.filterdb = filterdb
        self.selection_callback = selection_callback

        self.define_glade("define_filter", RULE_GLADE, also_load=["model1"])

        self.set_window(
            self.get_widget("define_filter"),
            self.get_widget("definition_title"),
            _("Define filter"),
        )
        self.setup_configs("interface.edit-filter", 500, 420)

        objectlist = self.get_widget("rule_list")
        self.rule_list = PersistentTreeView(uistate, "filt_rule")
        self.rule_list.set_vexpand(True)
        scrolledwindow = self.get_widget("scrolledwindow1")
        scrolledwindow.remove(objectlist)
        scrolledwindow.add(self.rule_list)

        self.rlist = ListModel(
            self.rule_list,
            [(_("Name"), -1, 150), (_("Values"), -1, 150)],
            self.select_row,
            self.on_edit_clicked,
        )

        self.fname = self.get_widget("filter_name")
        self.logical = self.get_widget("rule_apply")
        self.logical_not = self.get_widget("logical_not")
        self.comment = self.get_widget("comment")
        self.ok_btn = self.get_widget("definition_ok")
        self.edit_btn = self.get_widget("definition_edit")
        self.del_btn = self.get_widget("definition_delete")
        self.add_btn = self.get_widget("definition_add")

        self.ok_btn.connect("clicked", self.on_ok_clicked)
        self.edit_btn.connect("clicked", self.on_edit_clicked)
        self.del_btn.connect("clicked", self.on_delete_clicked)
        self.add_btn.connect("clicked", self.on_add_clicked)

        self.get_widget("definition_help").connect("clicked", self.on_help_clicked)
        self.get_widget("definition_cancel").connect("clicked", self.close_window)
        self.fname.connect("changed", self.filter_name_changed)

        op = self.filter.get_logical_op()
        # WARNING: must be listed in this order:
        self.logical.set_active(["and", "or", "one", "sequence"].index(op))
        self.logical_not.set_active(self.filter.get_invert())
        if self.filter.get_name():
            self.fname.set_text(self.filter.get_name())
        self.comment.set_text(self.filter.get_comment())
        self.draw_rules()

        self._set_size()
        self.rule_list.restore_column_size()
        self.show()

    def on_help_clicked(self, obj):
        """Display the relevant portion of Gramps manual"""
        display_help(webpage=WIKI_HELP_PAGE, section=WIKI_HELP_SEC2)

    def close_window(self, obj):
        self.close()

    def filter_name_changed(self, obj):
        name = str(self.fname.get_text())
        # Make sure that the name is not empty
        # and not in the list of existing filters (excluding this one)
        names = [
            filt.get_name()
            for filt in self.filterdb.get_filters(self.namespace)
            if filt != self.filter
        ]
        self.ok_btn.set_sensitive((len(name) != 0) and (name not in names))

    def select_row(self, obj):
        store, node = self.rlist.get_selected()
        if node:
            self.edit_btn.set_sensitive(True)
            self.del_btn.set_sensitive(True)
        else:
            self.edit_btn.set_sensitive(False)
            self.del_btn.set_sensitive(False)

    def draw_rules(self):
        self.rlist.clear()
        for r in self.filter.get_rules():
            self.rlist.add([r.name, r.display_values()], r)

    def on_ok_clicked(self, obj):
        n = str(self.fname.get_text()).strip()
        if n == "":
            return
        if n != self.filter.get_name():
            self.uistate.emit(
                "filter-name-changed", (self.namespace, str(self.filter.get_name()), n)
            )
        self.filter.set_name(n)
        self.filter.set_comment(str(self.comment.get_text()).strip())
        for f in self.filterdb.get_filters(self.namespace)[:]:
            if n == f.get_name():
                self.filterdb.get_filters(self.namespace).remove(f)
                break
        val = self.logical.get_active()
        # WARNING: must be listed in this order:
        op = (
            "and"
            if val == 0
            else "or" if val == 1 else "one" if val == 2 else "sequence"
        )
        self.logical.set_active(val)
        self.filter.set_logical_op(op)
        self.filter.set_invert(self.logical_not.get_active())
        self.filterdb.add(self.namespace, self.filter)
        if self.update:
            self.update()
        if self.selection_callback:
            self.selection_callback(self.filterdb, self.filter.get_name())
        self.close()

    def on_add_clicked(self, obj):
        try:
            EditRule(
                self.namespace,
                self.dbstate,
                self.uistate,
                self.track,
                self.filterdb,
                None,
                _("Add Rule"),
                self.update_rule,
                self.filter.get_name(),
            )
        except WindowActiveError:
            pass

    def on_edit_clicked(self, obj):
        store, node = self.rlist.get_selected()
        if node:
            d = self.rlist.get_object(node)

            try:
                EditRule(
                    self.namespace,
                    self.dbstate,
                    self.uistate,
                    self.track,
                    self.filterdb,
                    d,
                    _("Edit Rule"),
                    self.update_rule,
                    self.filter.get_name(),
                )
            except WindowActiveError:
                pass

    def update_rule(self, old_rule, new_rule):
        if old_rule is not None:
            self.filter.delete_rule(old_rule)
        self.filter.add_rule(new_rule)
        self.draw_rules()

    def on_delete_clicked(self, obj):
        store, node = self.rlist.get_selected()
        if node:
            gfilter = self.rlist.get_object(node)
            self.filter.delete_rule(gfilter)
            self.draw_rules()


# -------------------------------------------------------------------------
#
# ShowResults
#
# -------------------------------------------------------------------------
class ShowResults(ManagedWindow):
    def __init__(self, db, uistate, track, handle_list, filtname, namespace):
        ManagedWindow.__init__(self, uistate, track, self)

        self.db = db
        self.filtname = filtname
        self.namespace = namespace
        self.define_glade(
            "test",
            RULE_GLADE,
        )
        self.set_window(
            self.get_widget("test"), self.get_widget("test_title"), _("Filter Test")
        )
        self.setup_configs("interface.showresults", 450, 400)

        render = Gtk.CellRendererText()

        tree = self.get_widget("list")
        model = Gtk.ListStore(GObject.TYPE_STRING, GObject.TYPE_STRING)
        tree.set_model(model)

        column_n = Gtk.TreeViewColumn(_("Name"), render, text=0)
        tree.append_column(column_n)

        column_n = Gtk.TreeViewColumn(_("ID"), render, text=1)
        tree.append_column(column_n)

        self.get_widget("test_close").connect("clicked", self.close)

        new_list = sorted(
            (self.sort_val_from_handle(h) for h in handle_list),
            key=lambda x: glocale.sort_key(x[0]),
        )

        for s_, handle in new_list:
            name, gid = self.get_name_id(handle)
            model.append(row=[name, gid])

        self.show()

    def get_name_id(self, handle):
        if self.namespace == "Person":
            person = self.db.get_person_from_handle(handle)
            name = _nd.sorted(person)
            gid = person.get_gramps_id()
        elif self.namespace == "Family":
            family = self.db.get_family_from_handle(handle)
            name = family_name(family, self.db)
            gid = family.get_gramps_id()
        elif self.namespace == "Event":
            event = self.db.get_event_from_handle(handle)
            name = event.get_description()
            gid = event.get_gramps_id()
        elif self.namespace == "Source":
            source = self.db.get_source_from_handle(handle)
            name = source.get_title()
            gid = source.get_gramps_id()
        elif self.namespace == "Citation":
            citation = self.db.get_citation_from_handle(handle)
            name = citation.get_page()
            gid = citation.get_gramps_id()
        elif self.namespace == "Place":
            place = self.db.get_place_from_handle(handle)
            name = _pd.display(self.db, place)
            gid = place.get_gramps_id()
        elif self.namespace == "Media":
            obj = self.db.get_media_from_handle(handle)
            name = obj.get_description()
            gid = obj.get_gramps_id()
        elif self.namespace == "Repository":
            repo = self.db.get_repository_from_handle(handle)
            name = repo.get_name()
            gid = repo.get_gramps_id()
        elif self.namespace == "Note":
            note = self.db.get_note_from_handle(handle)
            name = note.get().replace("\n", " ")
            if len(name) > 80:
                name = name[:80] + "..."
            gid = note.get_gramps_id()
        return (name, gid)

    def sort_val_from_handle(self, handle):
        if self.namespace == "Person":
            name = self.db.get_person_from_handle(handle).get_primary_name()
            sortname = _nd.sort_string(name)
        elif self.namespace == "Family":
            sortname = family_name(self.db.get_family_from_handle(handle), self.db)
        elif self.namespace == "Event":
            sortname = self.db.get_event_from_handle(handle).get_description()
        elif self.namespace == "Source":
            sortname = self.db.get_source_from_handle(handle).get_title()
        elif self.namespace == "Citation":
            sortname = self.db.get_citation_from_handle(handle).get_page()
        elif self.namespace == "Place":
            place = self.db.get_place_from_handle(handle)
            sortname = _pd.display(self.db, place)
        elif self.namespace == "Media":
            sortname = self.db.get_media_from_handle(handle).get_description()
        elif self.namespace == "Repository":
            sortname = self.db.get_repository_from_handle(handle).get_name()
        elif self.namespace == "Note":
            gid = self.db.get_note_from_handle(handle).get_gramps_id()
            sortname = gid
        return (sortname, handle)


# -------------------------------------------------------------------------
#
# FilterEditor
#
# -------------------------------------------------------------------------
class FilterEditor(ManagedWindow):
    def __init__(self, namespace, filterdb, dbstate, uistate):
        ManagedWindow.__init__(self, uistate, [], FilterEditor)
        self.dbstate = dbstate
        self.db = dbstate.db
        self.filterdb = FilterList(filterdb)
        self.filterdb.load()
        self.namespace = namespace

        self.define_glade("filter_list", RULE_GLADE)
        self.edit = self.get_widget("filter_list_edit")
        self.clone = self.get_widget("filter_list_clone")
        self.delete = self.get_widget("filter_list_delete")
        self.test = self.get_widget("filter_list_test")

        self.edit.set_sensitive(False)
        self.clone.set_sensitive(False)
        self.delete.set_sensitive(False)
        self.test.set_sensitive(False)

        objectlist = self.get_widget("filters")
        self.filter_list = PersistentTreeView(self.uistate, "filt_list")
        scrolledwindow = self.get_widget("scrolledwindow2")
        scrolledwindow.remove(objectlist)
        scrolledwindow.add(self.filter_list)

        self.set_window(
            self.get_widget("filter_list"),
            self.get_widget("filter_list_title"),
            _TITLES[self.namespace],
        )
        self.setup_configs("interface.filter-editor", 400, 350)

        self.edit.connect("clicked", self.edit_filter)
        self.clone.connect("clicked", self.clone_filter)
        self.test.connect("clicked", self.test_clicked)
        self.delete.connect("clicked", self.delete_filter)

        self.connect_button("filter_list_help", self.help_clicked)
        self.connect_button("filter_list_close", self.close)
        self.connect_button("filter_list_add", self.add_new_filter)

        self.uistate.connect("filter-name-changed", self.clean_after_rename)

        self.clist = ListModel(
            self.filter_list,
            [(_("Filter"), 0, 150), (_("Comment"), 1, 150)],
            self.filter_select_row,
            self.edit_filter,
        )
        self.draw_filters()
        self._set_size()
        self.filter_list.restore_column_size()
        if not config.get("behavior.immediate-warn"):
            self.get_window().set_tooltip_text(_("Any changes are saved immediately"))
        self.show()

    def build_menu_names(self, obj):
        return (_("Custom Filter Editor"), _("Custom Filter Editor"))

    def help_clicked(self, obj):
        """Display the relevant portion of Gramps manual"""
        display_help(webpage=WIKI_HELP_PAGE, section=WIKI_HELP_SEC3)

    def filter_select_row(self, obj):
        store, node = self.clist.get_selected()
        if node:
            self.edit.set_sensitive(True)
            self.clone.set_sensitive(True)
            self.delete.set_sensitive(True)
            self.test.set_sensitive(True)
        else:
            self.edit.set_sensitive(False)
            self.clone.set_sensitive(False)
            self.delete.set_sensitive(False)
            self.test.set_sensitive(False)

    def close(self, *obj):
        self.filterdb.save()
        reload_custom_filters()
        # reload_system_filters()
        self.uistate.emit("filters-changed", (self.namespace,))
        ManagedWindow.close(self, *obj)

    def draw_filters(self):
        self.clist.clear()
        for f in self.filterdb.get_filters(self.namespace):
            self.clist.add([f.get_name(), f.get_comment()], f)

    def add_new_filter(self, obj):
        the_filter = GenericFilterFactory(self.namespace)()
        EditFilter(
            self.namespace,
            self.dbstate,
            self.uistate,
            self.track,
            the_filter,
            self.filterdb,
            self.draw_filters,
        )

    def edit_filter(self, obj):
        store, node = self.clist.get_selected()
        if node:
            gfilter = self.clist.get_object(node)
            EditFilter(
                self.namespace,
                self.dbstate,
                self.uistate,
                self.track,
                gfilter,
                self.filterdb,
                self.draw_filters,
            )

    def clone_filter(self, obj):
        store, node = self.clist.get_selected()
        if node:
            old_filter = self.clist.get_object(node)
            the_filter = GenericFilterFactory(self.namespace)(old_filter)
            the_filter.set_name("")
            EditFilter(
                self.namespace,
                self.dbstate,
                self.uistate,
                self.track,
                the_filter,
                self.filterdb,
                self.draw_filters,
            )

    def test_clicked(self, obj):
        store, node = self.clist.get_selected()
        if node:
            filt = self.clist.get_object(node)
            try:
                handle_list = filt.apply(self.db, self.get_all_handles())
            except FilterError as msg:
                (msg1, msg2) = msg.messages()
                ErrorDialog(msg1, msg2, parent=self.window)
                return
            ShowResults(
                self.db,
                self.uistate,
                self.track,
                handle_list,
                filt.get_name(),
                self.namespace,
            )

    def delete_filter(self, obj):
        store, node = self.clist.get_selected()
        if node:
            gfilter = self.clist.get_object(node)
            name = gfilter.get_name()
            if self.check_recursive_filters(self.namespace, name):
                QuestionDialog(
                    _("Delete Filter?"),
                    _(
                        "This filter is currently being used "
                        "as the base for other filters. Deleting "
                        "this filter will result in removing all "
                        "other filters that depend on it."
                    ),
                    _("Delete Filter"),
                    self._do_delete_selected_filter,
                    parent=self.window,
                )
            else:
                self._do_delete_selected_filter()

    def _do_delete_selected_filter(self):
        store, node = self.clist.get_selected()
        if node:
            gfilter = self.clist.get_object(node)
            self._do_delete_filter(self.namespace, gfilter)
            self.draw_filters()

    def _do_delete_filter(self, space, gfilter):
        # Find everything we need to remove
        filter_set = set()
        self._find_dependent_filters(space, gfilter, filter_set)

        # Remove what we found
        filters = self.filterdb.get_filters(space)
        list(map(filters.remove, filter_set))

    def _find_dependent_filters(self, space, gfilter, filter_set):
        """
        This method recursively calls itself to find all filters that
        depend on the given filter, either directly through one of the rules,
        or through the chain of dependencies.

        The filter_set is amended with the found filters.
        """
        filters = self.filterdb.get_filters(space)
        name = gfilter.get_name()
        for the_filter in filters:
            if the_filter.get_name() == name:
                continue
            for rule in the_filter.get_rules():
                values = list(rule.values())
                if issubclass(rule.__class__, MatchesFilterBase) and (name in values):
                    self._find_dependent_filters(space, the_filter, filter_set)
                    break
        # Add itself to the filter_set
        filter_set.add(gfilter)

    def get_all_handles(self):
        # Why use iter for some and get for others?
        if self.namespace == "Person":
            return self.db.iter_person_handles()
        elif self.namespace == "Family":
            return self.db.iter_family_handles()
        elif self.namespace == "Event":
            return self.db.get_event_handles()
        elif self.namespace == "Source":
            return self.db.get_source_handles()
        elif self.namespace == "Citation":
            return self.db.get_citation_handles()
        elif self.namespace == "Place":
            return self.db.iter_place_handles()
        elif self.namespace == "Media":
            return self.db.get_media_handles()
        elif self.namespace == "Repository":
            return self.db.get_repository_handles()
        elif self.namespace == "Note":
            return self.db.get_note_handles()

    def clean_after_rename(self, space, old_name, new_name):
        if old_name == "":
            return

        if old_name == new_name:
            return

        for the_filter in self.filterdb.get_filters(space):
            for rule in the_filter.get_rules():
                values = list(rule.values())
                if issubclass(rule.__class__, MatchesFilterBase) and (
                    old_name in values
                ):
                    ind = values.index(old_name)
                    values[ind] = new_name

    def check_recursive_filters(self, space, name):
        for the_filter in self.filterdb.get_filters(space):
            for rule in the_filter.get_rules():
                values = list(rule.values())
                if issubclass(rule.__class__, MatchesFilterBase) and (name in values):
                    return True
        return False



#! /usr/bin/env python3
"""
Gramps - a GTK+/GNOME based genealogy program

Copyright (c) 2016 Gramps Development Team

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
"""

import unittest

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.utils.id import set_det_id
from gramps.gen.db import DbTxn
from gramps.gen.db.utils import make_database
from gramps.gen.dbstate import DbState
from gramps.cli.clidbman import CLIDbManager
from gramps.gen.lib import Person, Family, Note, Surname
from gramps.gen.merge import MergeFamilyQuery
from gramps.gen.utils.callman import CallbackManager


class DbTestClassBase(object):
    # Inherit from `object` so unittest doesn't think these are tests which
    # should be run

    @classmethod
    def setUpClass(cls):
        # print("doing setup with", cls.param)
        set_det_id(True)  # handles must be deterministic for this test
        cls.dbstate = DbState()
        cls.dbman = CLIDbManager(cls.dbstate)
        dirpath, _name = cls.dbman.create_new_db_cli(
            "Test: %s" % cls.param, dbid=cls.param
        )
        cls.db = make_database(cls.param)
        cls.db.load(dirpath, None)

    @classmethod
    def tearDownClass(cls):
        # print("tear down", cls.param)
        cls.db.close()
        cls.dbman.remove_database("Test: %s" % cls.param)

    def __add_person(self, gender, first_name, surname, trans):
        person = Person()
        person.gender = gender
        _name = person.primary_name
        _name.first_name = first_name
        surname1 = Surname()
        surname1.surname = surname
        _name.set_surname_list([surname1])
        self.db.add_person(person, trans)
        return person

    def __add_note(self, text, trans):
        note = Note(text=text)
        self.db.add_note(note, trans)
        return note

    def __add_family(self, father, mother, trans):
        family = Family()
        family.set_father_handle(father.handle)
        family.set_mother_handle(mother.handle)
        fam_handle = self.db.add_family(family, trans)
        father.add_family_handle(fam_handle)
        mother.add_family_handle(fam_handle)
        self.db.commit_person(father, trans)
        self.db.commit_person(mother, trans)
        return family

    def __setup_callbacks(self):
        self.db.connect("family-add", self._family_add)
        self.db.connect("family-update", self._family_update)
        self.db.connect("family-delete", self._family_delete)
        self.db.connect("person-add", self._person_add)
        self.db.connect("person-update", self._person_update)
        self.db.connect("person-delete", self._person_delete)
        self.db.connect("note-add", self._note_add)
        self.db.connect("note-update", self._note_update)
        self.db.connect("note-delete", self._note_delete)
        # we will also test the CallbackManager by watching a single person
        self.cm_sigs = []
        self.callman = CallbackManager(self.db)
        self.callman.register_callbacks(
            {
                "person-delete": self._cm_pers_delete,
                "person-update": self._cm_pers_update,
                "person-add": self._cm_pers_add,
            }
        )
        self.callman.connect_all(keys=["person"])
        self.callman.register_handles(
            {"person": ["0000000300000003", "0000000700000007"]}
        )

    def _family_add(self, *args):
        self._log_sig("family-add", args)

    def _family_update(self, *args):
        self._log_sig("family-update", args)

    def _family_delete(self, *args):
        self._log_sig("family-delete", args)

    def _person_add(self, *args):
        self._log_sig("person-add", args)

    def _person_update(self, *args):
        self._log_sig("person-update", args)

    def _person_delete(self, *args):
        self._log_sig("person-delete", args)

    def _note_add(self, *args):
        self._log_sig("note-add", args)

    def _note_update(self, *args):
        self._log_sig("note-update", args)

    def _note_delete(self, *args):
        self._log_sig("note-delete", args)

    def _log_sig(self, sig, args):
        # print("('%s', %s)," % (sig, args))
        self.sigs.append((sig, args[0]))

    def _cm_pers_add(self, *args):
        self.cm_sigs.append(("person-add", args[0]))

    def _cm_pers_update(self, *args):
        self.cm_sigs.append(("person-update", args[0]))

    def _cm_pers_delete(self, *args):
        self.cm_sigs.append(("person-delete", args[0]))

    def test_one(self):
        self.__setup_callbacks()
        self.sigs = []
        with DbTxn("Add test objects", self.db) as trans:
            father1 = self.__add_person(Person.MALE, "John", "Allen", trans)
            mother1 = self.__add_person(Person.FEMALE, "Mary", "Allen", trans)
            father2 = self.__add_person(Person.MALE, "John", "Baker", trans)
            mother2 = self.__add_person(Person.FEMALE, "Mary", "Baker", trans)
            family1 = self.__add_family(father1, mother1, trans)
            family2 = self.__add_family(father2, mother2, trans)
        self.callman.register_obj(father2, directonly=True)
        self.callman.register_obj(father2, directonly=False)
        sigs = [
            (
                "person-add",
                [
                    "0000000100000001",
                    "0000000200000002",
                    "0000000300000003",
                    "0000000400000004",
                ],
            ),
            ("family-add", ["0000000500000005", "0000000600000006"]),
            (
                "person-update",
                [
                    "0000000100000001",
                    "0000000200000002",
                    "0000000300000003",
                    "0000000400000004",
                ],
            ),
        ]
        self.assertEqual(sigs, self.sigs, msg="make families")
        # save state for later undo/redo check
        step1 = (family1, father1, mother1, family2, father2, mother2)
        fam_cnt = self.db.get_number_of_families()
        pers_cnt = self.db.get_number_of_people()
        self.assertEqual(fam_cnt, 2, msg="make families check")
        self.assertEqual(pers_cnt, 4, msg="make families persons check")

        # lets do a family merge, this will not only combine the families, but
        # also combine the the fathers and mothers as well
        self.sigs = []
        query = MergeFamilyQuery(
            self.db, family1, family2, father1.handle, mother1.handle
        )
        query.execute()
        sigs = [
            ("person-delete", ["0000000300000003", "0000000400000004"]),
            ("family-delete", ["0000000600000006"]),
            (
                "person-update",
                [
                    "0000000100000001",
                    "0000000200000002",
                    "0000000100000001",
                    "0000000200000002",
                ],
            ),
            ("family-update", ["0000000500000005", "0000000500000005"]),
        ]
        self.assertEqual(sigs, self.sigs, msg="merge families")
        fam_cnt = self.db.get_number_of_families()
        pers_cnt = self.db.get_number_of_people()
        self.assertEqual(fam_cnt, 1, msg="merge families check")
        self.assertEqual(pers_cnt, 2, msg="merge families persons check")
        # save new family'people for later redo check
        family1 = self.db.get_family_from_handle(family1.handle)
        father1 = self.db.get_person_from_handle(father1.handle)
        mother1 = self.db.get_person_from_handle(mother1.handle)
        step2 = (family1, father1, mother1)

        # we check that update and add signals are not emitted if the same
        # object is deleted in the same transation
        self.sigs = []
        with DbTxn("Note add/update/delete", self.db) as trans:
            note = self.__add_note("some text", trans)
            note.set("some other text")
            self.db.commit_note(note, trans)
            self.db.remove_note(note.handle, trans)
        sigs = [("note-delete", ["0000000700000007"])]
        self.assertEqual(sigs, self.sigs, msg="note signals check")
        note_cnt = self.db.get_number_of_notes()
        self.assertEqual(note_cnt, 0, msg="note check")

        # Test some undos, start with the note undo
        self.sigs = []
        self.db.undo()
        sigs = [("note-delete", ["0000000700000007"])]
        self.assertEqual(sigs, self.sigs, msg="undo note signals check")

        # Test merge undo
        self.sigs = []
        self.db.undo()
        sigs = [
            ("person-add", ["0000000400000004", "0000000300000003"]),
            ("family-add", ["0000000600000006"]),
            (
                "person-update",
                [
                    "0000000200000002",
                    "0000000100000001",
                    "0000000200000002",
                    "0000000100000001",
                ],
            ),
            (
                "family-update",
                [
                    "0000000500000005",
                    "0000000600000006",
                    "0000000600000006",
                    "0000000500000005",
                ],
            ),
        ]
        self.assertEqual(sigs, self.sigs, msg="undo merge signals check")
        fam_cnt = self.db.get_number_of_families()
        pers_cnt = self.db.get_number_of_people()
        self.assertEqual(fam_cnt, 2, msg="undo merge families check")
        self.assertEqual(pers_cnt, 4, msg="undo merge families persons check")
        # step1 = (family1, father1, mother1, family2, father2, mother2)
        obj_s = self.db.get_family_from_handle(family1.handle).serialize()
        self.assertEqual(
            obj_s, step1[0].serialize(), msg="undo merge families fam1 check"
        )
        obj_s = self.db.get_person_from_handle(father1.handle).serialize()
        self.assertEqual(
            obj_s, step1[1].serialize(), msg="undo merge families father1 check"
        )
        obj_s = self.db.get_person_from_handle(mother1.handle).serialize()
        self.assertEqual(
            obj_s, step1[2].serialize(), msg="undo merge families mother1 check"
        )
        obj_s = self.db.get_family_from_handle(family2.handle).serialize()
        self.assertEqual(
            obj_s, step1[3].serialize(), msg="undo merge families fam2 check"
        )
        obj_s = self.db.get_person_from_handle(father2.handle).serialize()
        self.assertEqual(
            obj_s, step1[4].serialize(), msg="undo merge families father2 check"
        )
        obj_s = self.db.get_person_from_handle(mother2.handle).serialize()
        self.assertEqual(
            obj_s, step1[5].serialize(), msg="undo merge families mother2 check"
        )

        # Test family build undo
        self.sigs = []
        self.db.undo()
        sigs = [
            (
                "person-delete",
                [
                    "0000000400000004",
                    "0000000300000003",
                    "0000000200000002",
                    "0000000100000001",
                ],
            ),
            ("family-delete", ["0000000600000006", "0000000500000005"]),
        ]
        self.assertEqual(sigs, self.sigs, msg="undo family build signals check")
        fam_cnt = self.db.get_number_of_families()
        pers_cnt = self.db.get_number_of_people()
        self.assertEqual(fam_cnt, 0, msg="undo make families check")
        self.assertEqual(pers_cnt, 0, msg="undo make families persons check")

        # Test family build redo
        self.sigs = []
        self.db.redo()
        sigs = [
            (
                "person-add",
                [
                    "0000000100000001",
                    "0000000200000002",
                    "0000000300000003",
                    "0000000400000004",
                ],
            ),
            ("family-add", ["0000000500000005", "0000000600000006"]),
            (
                "person-update",
                [
                    "0000000100000001",
                    "0000000200000002",
                    "0000000300000003",
                    "0000000400000004",
                ],
            ),
        ]
        self.assertEqual(sigs, self.sigs, msg="redo family build signals check")
        fam_cnt = self.db.get_number_of_families()
        pers_cnt = self.db.get_number_of_people()
        self.assertEqual(fam_cnt, 2, msg="redo make families check")
        self.assertEqual(pers_cnt, 4, msg="redo make families persons check")
        # step1 = (family1, father1, mother1, family2, father2, mother2)
        obj_s = self.db.get_family_from_handle(family1.handle).serialize()
        self.assertEqual(
            obj_s, step1[0].serialize(), msg="redo merge families fam1 check"
        )
        obj_s = self.db.get_person_from_handle(father1.handle).serialize()
        self.assertEqual(
            obj_s, step1[1].serialize(), msg="redo merge families father1 check"
        )
        obj_s = self.db.get_person_from_handle(mother1.handle).serialize()
        self.assertEqual(
            obj_s, step1[2].serialize(), msg="redo merge families mother1 check"
        )
        obj_s = self.db.get_family_from_handle(family2.handle).serialize()
        self.assertEqual(
            obj_s, step1[3].serialize(), msg="redo merge families fam2 check"
        )
        obj_s = self.db.get_person_from_handle(father2.handle).serialize()
        self.assertEqual(
            obj_s, step1[4].serialize(), msg="redo merge families father2 check"
        )
        obj_s = self.db.get_person_from_handle(mother2.handle).serialize()
        self.assertEqual(
            obj_s, step1[5].serialize(), msg="redo merge families mother2 check"
        )

        # Test family merge redo
        self.sigs = []
        self.db.redo()
        sigs = [
            ("person-delete", ["0000000300000003", "0000000400000004"]),
            ("family-delete", ["0000000600000006"]),
            (
                "person-update",
                [
                    "0000000100000001",
                    "0000000200000002",
                    "0000000100000001",
                    "0000000200000002",
                ],
            ),
            ("family-update", ["0000000500000005", "0000000500000005"]),
        ]
        self.assertEqual(sigs, self.sigs, msg="merge families")
        fam_cnt = self.db.get_number_of_families()
        pers_cnt = self.db.get_number_of_people()
        self.assertEqual(fam_cnt, 1, msg="merge families check")
        self.assertEqual(pers_cnt, 2, msg="merge families persons check")
        # step2 = (family1, father1, mother1)
        obj_s = self.db.get_family_from_handle(family1.handle).serialize()
        self.assertEqual(
            obj_s, step2[0].serialize(), msg="undo merge families fam1 check"
        )
        obj_s = self.db.get_person_from_handle(father1.handle).serialize()
        self.assertEqual(
            obj_s, step2[1].serialize(), msg="undo merge families father1 check"
        )
        obj_s = self.db.get_person_from_handle(mother1.handle).serialize()
        self.assertEqual(
            obj_s, step2[2].serialize(), msg="undo merge families mother1 check"
        )

        # Test note redo
        self.sigs = []
        self.db.redo()
        sigs = [("note-delete", ["0000000700000007"])]
        self.assertEqual(sigs, self.sigs, msg="undo note signals check")

        # now lets see if the callback manager is doing its job
        # print(self.cm_sigs)
        sigs = [
            ("person-add", ["0000000300000003"]),
            ("person-update", ["0000000300000003"]),
            ("person-delete", ["0000000300000003"]),
            ("person-add", ["0000000300000003"]),
            ("person-delete", ["0000000300000003"]),
            ("person-add", ["0000000300000003"]),
            ("person-update", ["0000000300000003"]),
            ("person-delete", ["0000000300000003"]),
        ]

        self.assertEqual(sigs, self.cm_sigs, msg="callback manager signals check")
        self.callman.unregister_handles({"person": ["0000000300000003"]})
        # we have to look deep into cm to see if handles are really deleted
        cm_persons = self.callman._CallbackManager__handles["person"]
        self.assertEqual(
            cm_persons, ["0000000700000007"], msg="Callback Manager unregister check"
        )
        self.callman.unregister_all()
        # we have to look deep into cm to see if handles are really deleted
        cm_persons = self.callman._CallbackManager__handles["person"]
        self.assertEqual(cm_persons, [], msg="Callback Manager unregister check")
        # we have to look deep into cm to see if callbacks are really deleted
        cm_padd_key = self.callman._CallbackManager__callbacks["person-add"][1]
        self.assertEqual(cm_padd_key, 10, msg="Callback Manager cb check")
        self.callman.disconnect_all()
        # we have to look deep into cm to see if callbacks are really deleted
        cm_padd_key = self.callman._CallbackManager__callbacks["person-add"][1]
        self.assertEqual(cm_padd_key, None, msg="Callback Manager disconnect cb check")


class TestSQLite(DbTestClassBase, unittest.TestCase):
    param = "sqlite"


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2007       Johan Gonqvist <johan.gronqvist@gmail.com>
# Copyright (C) 2007-2009  Gary Burton <gary.burton@zen.co.uk>
# Copyright (C) 2007-2009  Stephane Charette <stephanecharette@gmail.com>
# Copyright (C) 2008-2009  Brian G. Matherly
# Copyright (C) 2008       Jason M. Simanek <jason@bohemianalps.com>
# Copyright (C) 2008-2011  Rob G. Healey <robhealey1@gmail.com>
# Copyright (C) 2010       Doug Blank <doug.blank@gmail.com>
# Copyright (C) 2010       Jakim Friant
# Copyright (C) 2010-      Serge Noiraud
# Copyright (C) 2011       Tim G L Lyons
# Copyright (C) 2013       Benny Malengier
# Copyright (C) 2016       Allen Crider
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Narrative Web Page generator.

Classe:
    EventPage - Event index page and individual Event pages
"""
# ------------------------------------------------
# python modules
# ------------------------------------------------
from collections import defaultdict
from operator import itemgetter
from decimal import getcontext
import unicodedata
import logging

# ------------------------------------------------
# Gramps module
# ------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale
from gramps.gen.lib import Date, Event
from gramps.gen.plug.report import Bibliography
from gramps.plugins.lib.libhtml import Html

# ------------------------------------------------
# specific narrative web import
# ------------------------------------------------
from gramps.plugins.webreport.basepage import BasePage
from gramps.plugins.webreport.common import (
    _EVENTMAP,
    alphabet_navigation,
    partial_navigation,
    create_indexes_pages,
    FULLCLEAR,
    sort_event_types,
    AlphabeticIndex,
)

_ = glocale.translation.sgettext
LOG = logging.getLogger(".NarrativeWeb")
getcontext().prec = 8


#################################################
#
#    creates the Event List Page and EventPages
#
#################################################
class EventPages(BasePage):
    """
    This class is responsible for displaying information about the 'Person'
    database objects. It displays this information under the 'Events'
    tab. It is told by the 'add_instances' call which 'Person's to display,
    and remembers the list of persons. A single call to 'display_pages'
    displays both the Event List (Index) page and all the Event
    pages.

    The base class 'BasePage' is initialised once for each page that is
    displayed.
    """

    def __init__(self, report, the_lang, the_title):
        """
        @param: report    -- The instance of the main report class
                             for this report
        @param: the_lang  -- The lang to process
        @param: the_title -- The title page related to the language
        """
        BasePage.__init__(self, report, the_lang, the_title)
        self.event_handle_list = []
        self.event_types = []
        self.event_dict = defaultdict(set)

    def display_pages(self, the_lang, the_title):
        """
        Generate and output the pages under the Event tab, namely the event
        index and the individual event pages.

        @param: the_lang  -- The lang to process
        @param: the_title -- The title page related to the language
        """
        BasePage.__init__(self, self.report, the_lang, the_title)
        LOG.debug("obj_dict[Event]")
        for item in self.report.obj_dict[Event].items():
            LOG.debug("    %s", str(item))
        event_handle_list = self.report.obj_dict[Event].keys()
        event_types = []
        for event_handle in event_handle_list:
            event = self.r_db.get_event_from_handle(event_handle)
            event_types.append(self._(event.get_type().xml_str()))
        message = _("Creating event pages")
        progress_title = self.report.pgrs_title(the_lang)
        with self.r_user.progress(
            progress_title, message, len(event_handle_list) + 1
        ) as step:
            index = 1
            for event_handle in event_handle_list:
                step()
                index += 1
                self.eventpage(self.report, the_lang, the_title, event_handle)
            step()
        self.eventlistpage(
            self.report, the_lang, the_title, event_types, event_handle_list
        )

    def __output_event(
        self,
        ldatec,
        event_type,
        tbody,
        bucket_letter,
        first_letter,
        _event_displayed,
        first_type,
        event_handle,
    ):
        """
        Generate and output the data for a single event

        @param: ldatec          -- Last change date and time (updated)
        @param: event_type      -- The event type name processed for XML etc
        @param: tbody           -- The current HTML body into which the data is
                                   assembled
        @param: bucket_letter   -- The AlphabeticIndex bucket for this event
        @param: first_letter    -- Whether this is the first event for this
                                   letter
        @param: event_displayed -- List of events already displayed
        @param: first_type      -- Whether this is the first event of this type
        @param: event_handle    -- Handle of the event to be output

        :returns: Returns a tuple of updated ldatec, first_letter, first_type,
                    _event_displayed
        :rtype: tuple

        """
        event = self.r_db.get_event_from_handle(event_handle)
        _type = event.get_type()
        gid = event.get_gramps_id()
        if event.get_change_time() > ldatec:
            ldatec = event.get_change_time()
        # check to see if we have listed this gramps_id yet?
        if gid not in _event_displayed:
            if int(_type) in _EVENTMAP:
                handle_list = set(
                    self.r_db.find_backlink_handles(
                        event_handle, include_classes=["Family", "Person"]
                    )
                )
            else:
                handle_list = set(
                    self.r_db.find_backlink_handles(
                        event_handle, include_classes=["Person"]
                    )
                )
            if handle_list:
                trow = Html("tr")
                tbody += trow
                # set up hyperlinked letter for alphabet_navigation
                tcell = Html("td", class_="ColumnLetter", inline=True)
                trow += tcell
                if first_letter:
                    first_letter = False
                    first_type = False
                    # Update the ColumnLetter cell and
                    # create a populated ColumnType
                    t_a = 'class = "BeginLetter BeginType"'
                    trow.attr = t_a
                    letter = bucket_letter
                    ttle = self._("Event types beginning " "with letter %s") % letter
                    tcell += Html("a", letter, name=letter, title=ttle)
                    tcell = Html(
                        "td", class_="ColumnType", title=self._(event_type), inline=True
                    )
                    tcell += self._(event_type)
                elif first_type:
                    first_type = False
                    # Update the ColumnLetter cell and
                    # create a populated ColumnType cell
                    trow.attr = 'class = "BeginType"'
                    tcell = Html(
                        "td", class_="ColumnType", title=self._(event_type), inline=True
                    )
                    tcell += self._(event_type)
                else:
                    tcell = Html("td", class_="ColumnType", title="&nbsp;", inline=True)
                    tcell += "&nbsp;"  # create a non-populated ColumnType
                # Add the ColumnType cell
                trow += tcell
                # event date
                tcell = Html("td", class_="ColumnDate", inline=True)
                trow += tcell
                date = Date.EMPTY
                if event:
                    date = event.get_date_object()
                    if date and date is not Date.EMPTY:
                        tcell += self.rlocale.get_date(date)
                else:
                    tcell += "&nbsp;"
                # Gramps ID
                trow += Html("td", class_="ColumnGRAMPSID") + (
                    self.event_grampsid_link(event_handle, gid, None)
                )
                # Person(s) column
                tcell = Html("td", class_="ColumnPerson")
                trow += tcell
                # classname can either be a person or a family
                first_person = True
                # get person(s) for ColumnPerson
                sorted_list = sorted(handle_list)
                self.complete_people(tcell, first_person, sorted_list, uplink=False)
        _event_displayed.append(gid)
        return (ldatec, first_letter, first_type, _event_displayed)

    def part_eventlistpage(
        self,
        report,
        index_list,
        name,
        letter,
        index,
        part=None,
        subp=None,
        partial_list=None,
    ):
        """
        Will create the part of event list page
        """
        nav_name = part_name = name
        if part != 0 and subp != 0:
            part_name += str(part)
            nav_name = part_name
        if subp and subp != 0:
            part_name += subp
        output_file, sio = self.report.create_file(part_name)
        result = self.write_header(self._("Events"))
        eventslistpage, dummy_head, dummy_body, outerwrapper = result
        ldatec = 0

        # begin events list  division
        with Html("div", class_="content", id="EventList") as eventlist:
            outerwrapper += eventlist

            msg = self._(
                "This page contains an index of all the events in the "
                "database, sorted by their type and date (if one is "
                "present). Clicking on an event&#8217;s Gramps ID "
                "will open a page for that event."
            )
            eventlist += Html("p", msg, id="description")

            # Output the navigation
            alpha_nav = alphabet_navigation(
                index_list,
                self.rlocale,
                current=part_name,
                rtl=self.dir,
                new_page="events",
                ext=self.ext,
            )
            if alpha_nav:
                eventlist += alpha_nav

            if part is not None:
                # We need to create a new navigation tab for partial page index.
                partial_nav = partial_navigation(
                    partial_list,
                    self.rlocale,
                    current=part_name,
                    rtl=self.dir,
                    ext=self.ext,
                    new_page=nav_name,
                )
                if partial_nav is not None:
                    eventlist += Html(
                        "div", style="clear: both;"
                    )  # This to align the next div to the left.
                    eventlist += partial_nav

            # begin alphabet event table
            with Html(
                "table", class_="infolist primobjlist alphaevent " + self.dir
            ) as table:
                eventlist += table

                thead = Html("thead")
                table += thead

                trow = Html("tr")
                thead += trow

                trow.extend(
                    Html("th", label, class_=colclass, inline=True)
                    for (label, colclass) in [
                        (self._("Letter"), "ColumnRowLabel"),
                        (self._("Type"), "ColumnType"),
                        (self._("Date"), "ColumnDate"),
                        (self._("Gramps ID"), "ColumnGRAMPSID"),
                        (self._("Person"), "ColumnPerson"),
                    ]
                )

                tbody = Html("tbody")
                table += tbody

                first_letter = True
                first_type = True
                _event_displayed = []
                # for each bucket, output the events in that bucket
                # letter, handle_list = index
                evttype = ""
                for evt_hdle, event_type in index:
                    evt_type = event_type.split(" (")[0]
                    if evt_type != evttype:
                        evttype = evt_type
                        first_type = True
                    (
                        ldatec,
                        first_letter,
                        first_type,
                        _event_displayed,
                    ) = self.__output_event(
                        ldatec,
                        evt_type,
                        tbody,
                        letter,
                        first_letter,
                        _event_displayed,
                        first_type,
                        evt_hdle,
                    )

        # add clearline for proper styling
        # add footer section
        footer = self.write_footer(ldatec)
        outerwrapper += (FULLCLEAR, footer)

        # send page ut for processing
        # and close the file
        self.xhtml_writer(eventslistpage, output_file, sio, ldatec)

    def eventlistpage(
        self, report, the_lang, the_title, event_types, event_handle_list
    ):
        """
        Will create the event list page

        @param: report            -- The instance of the main report class for
                                     this report
        @param: the_lang          -- The lang to process
        @param: the_title         -- The title page related to the language
        @param: event_types       -- A list of the type in the events database
        @param: event_handle_list -- A list of event handles
        """
        BasePage.__init__(self, report, the_lang, the_title)
        # get alphabet navigation...
        # Assemble all the event types
        index = AlphabeticIndex(self.rlocale)
        for event_type, data_list in sort_event_types(
            self.r_db, event_types, event_handle_list, self.rlocale
        ):
            index.addRecord(event_type, data_list)

        # Extract the buckets from the index
        index_list = []
        index.resetBucketIterator()
        events_handle_dict = defaultdict(list)
        events_handle_list = []
        max_letter_rows = defaultdict(int)
        while index.nextBucket():
            if index.bucketRecordCount != 0:
                letter = index.bucketLabel
                if letter not in index_list:
                    index_list.append(letter)
                bletter = (
                    unicodedata.normalize("NFKD", letter)[0]
                    if len(letter) > 0
                    else letter
                )
                while index.nextRecord():
                    handle_list = index.recordData
                    if handle_list:
                        for handle in handle_list:
                            event = self.r_db.get_event_from_handle(handle[1])
                            if event:
                                date = event.get_date_object()
                                sdate = str(self._geteventdate(event))
                                evt_date = ""
                                if date and date is not Date.EMPTY:
                                    evt_date = self.rlocale.get_date(date)
                                if evt_date == "":
                                    evt_date = event.get_gramps_id()
                                goto = index.recordName + " (%s)" % evt_date
                            events_handle_dict[bletter].append((handle[1], sdate, goto))
                        if bletter in max_letter_rows:
                            max_letter_rows[bletter] += len(handle_list)
                        else:
                            max_letter_rows[bletter] = len(handle_list)
        events_handle_list = list(events_handle_dict.items())
        extended_handle_list = defaultdict(list)
        for bletter, hdlel in events_handle_list:
            bletter = (
                unicodedata.normalize("NFKD", bletter)[0]
                if len(bletter) > 0
                else bletter
            )
            bletter = bletter[0] if len(bletter) > 0 else bletter
            bletter = bletter.upper() if bletter.isalpha() else "…"

            def sort_on_event_and_date(evt):
                return evt[2].split(" (")[0], evt[1]

            hdlel = sorted(hdlel, key=sort_on_event_and_date)
            for hdle in hdlel:
                extended_handle_list[bletter].append((hdle[0], hdle[2]))
        extended_handle_list = list(extended_handle_list.items())
        create_indexes_pages(
            report,
            "events",
            index_list,
            self.part_eventlistpage,
            extended_handle_list,
            locale=self.rlocale,
        )

    def _geteventdate(self, event):
        """
        Get the event date

        @param: event_handle -- The handle for the event to use
        """
        event_date = Date.EMPTY
        if event:
            date = event.get_date_object()
            if date:
                # returns the date in YYYY-MM-DD format
                return Date(
                    date.get_year_calendar("Gregorian"),
                    date.get_month(),
                    date.get_day(),
                )

        # return empty date string
        return event_date

    def event_grampsid_link(self, handle, grampsid, uplink):
        """
        Create a hyperlink from event handle, but show grampsid

        @param: handle   -- The handle for the event
        @param: grampsid -- The gramps ID to display
        @param: uplink   -- If True, then "../../../" is inserted in front of
                            the result.
        """
        url = self.report.build_url_fname_html(handle, "evt", uplink)

        # return hyperlink to its caller
        return Html("a", grampsid, href=url, title=grampsid, inline=True)

    def eventpage(self, report, the_lang, the_title, event_handle):
        """
        Creates the individual event page

        @param: report       -- The instance of the main report class for
                                this report
        @param: the_lang     -- The lang to process
        @param: the_title    -- The title page related to the language
        @param: event_handle -- The event handle for the database
        """
        event = report.database.get_event_from_handle(event_handle)
        BasePage.__init__(self, report, the_lang, the_title, event.get_gramps_id())
        if not event:
            return

        ldatec = event.get_change_time()
        event_media_list = event.get_media_list()
        self.uplink = True
        subdirs = True
        evt_type = self._(event.get_type().xml_str())
        self.page_title = "%(eventtype)s" % {"eventtype": evt_type}
        self.bibli = Bibliography()

        output_file, sio = self.report.create_file(event_handle, "evt")
        result = self.write_header(self._("Events"))
        eventpage, head, dummy_body, outerwrapper = result
        if event_media_list and self.create_media:
            if self.the_lang and not self.usecms:
                fname = "/".join(["..", "css", "lightbox.css"])
                jsname = "/".join(["..", "css", "lightbox.js"])
            else:
                fname = "/".join(["css", "lightbox.css"])
                jsname = "/".join(["css", "lightbox.js"])
            url = self.report.build_url_fname(fname, None, self.uplink)
            head += Html(
                "link", href=url, type="text/css", media="screen", rel="stylesheet"
            )
            url = self.report.build_url_fname(jsname, None, self.uplink)
            head += Html("script", src=url, type="text/javascript", inline=True)

        # start event detail division
        with Html("div", class_="content", id="EventDetail") as eventdetail:
            outerwrapper += eventdetail

            thumbnail = self.disp_first_img_as_thumbnail(event_media_list, event)
            if thumbnail is not None:
                eventdetail += thumbnail

            # display page title
            eventdetail += Html("h3", self.page_title, inline=True)

            # begin eventdetail table
            with Html("table", class_="infolist eventlist") as table:
                eventdetail += table

                tbody = Html("tbody")
                table += tbody

                evt_gid = event.get_gramps_id()
                if not self.noid and evt_gid:
                    trow = Html("tr") + (
                        Html(
                            "td",
                            self._("Gramps ID"),
                            class_="ColumnAttribute",
                            inline=True,
                        ),
                        Html("td", evt_gid, class_="ColumnGRAMPSID", inline=True),
                    )
                    tbody += trow

                # get event data
                #
                # for more information: see get_event_data()
                #
                event_data = self.get_event_data(event, event_handle, subdirs, evt_gid)

                for label, colclass, data in event_data:
                    if data:
                        trow = Html("tr") + (
                            Html("td", label, class_="ColumnAttribute", inline=True),
                            Html("td", data, class_="Column" + colclass),
                        )
                        tbody += trow

                # Tags
                tags = self.show_tags(event)
                if tags and self.report.inc_tags:
                    trow = Html("tr") + (
                        Html(
                            "td", self._("Tags"), class_="ColumnAttribute", inline=True
                        ),
                        Html("td", tags, class_="ColumnValue", inline=True),
                    )
                    table += trow

            # Narrative subsection
            notelist = event.get_note_list()
            notelist = self.display_note_list(notelist, Event)
            if notelist is not None:
                eventdetail += notelist

            # get attribute list
            attrlist = event.get_attribute_list()
            if attrlist:
                attrsection, attrtable = self.display_attribute_header()
                self.display_attr_list(attrlist, attrtable)
                eventdetail += attrsection

            # event source references
            srcrefs = self.display_ind_sources(event)
            if srcrefs is not None:
                eventdetail += srcrefs

            # display additional images as gallery
            if self.create_media:
                addgallery = self.disp_add_img_as_gallery(event_media_list, event)
                if addgallery:
                    eventdetail += addgallery

            # References list
            ref_list = self.display_bkref_list(Event, event_handle)
            if ref_list is not None:
                eventdetail += ref_list

        # add clearline for proper styling
        # add footer section
        footer = self.write_footer(ldatec)
        outerwrapper += (FULLCLEAR, footer)

        # send page out for processing
        # and close the page
        self.xhtml_writer(eventpage, output_file, sio, ldatec)
